import dataclasses
from typing import Dict, List

from pynmrstar import Loop, Saveframe

from nef_pipelines.lib.nef_lib import UNUSED, loop_row_namespace_iter
from nef_pipelines.lib.structures import (
    AtomLabel,
    Residue,
    SequenceResidue,
    ShiftData,
    ShiftList,
)

NEF_CHEMICAL_SHIFT_LOOP = "nef_chemical_shift"


def nef_frames_to_shifts(frames: List[Saveframe]) -> List[ShiftData]:
    """

    :param frames:
    :return:
    """
    shifts = []

    residue_field_names = [field.name for field in dataclasses.fields(SequenceResidue)]
    atom_field_names = [
        field.name for field in dataclasses.fields(AtomLabel) if field.name != "residue"
    ]
    for frame in frames:

        loop = frame.get_loop(NEF_CHEMICAL_SHIFT_LOOP)

        for row in loop_row_namespace_iter(loop):

            residue_fields = {
                name: value
                for name, value in vars(row).items()
                if name in residue_field_names
            }
            atom_fields = {
                name: value
                for name, value in vars(row).items()
                if name in atom_field_names
            }
            residue = Residue(**residue_fields)
            label = AtomLabel(residue, **atom_fields)
            shifts.append(ShiftData(label, row.value, row.value_uncertainty))

    return shifts


def shifts_to_nef_frame(shift_list: ShiftList, frame_name: str):
    """
    convert a shift list to a nef chemical shift list frame
    :param shift_list: the shifts
    :param frame_name: the name for the frame to appended to nef_chemical_shift_list
    :return: the shift list frame
    """
    SHIFT_LIST_FRAME_CATEGORY = "nef_chemical_shift_list"
    SHIFT_LOOP_CATEGORY = "nef_chemical_shift"

    frame_code = f"{SHIFT_LIST_FRAME_CATEGORY}_{frame_name}"

    frame = Saveframe.from_scratch(frame_code, SHIFT_LIST_FRAME_CATEGORY)

    frame.add_tag("sf_category", SHIFT_LIST_FRAME_CATEGORY)
    frame.add_tag("sf_framecode", frame_code)

    loop = Loop.from_scratch()
    frame.add_loop(loop)

    tags = (
        "chain_code",
        "sequence_code",
        "residue_name",
        "atom_name",
        "value",
        "value_uncertainty",
        "element",
        "isotope_number",
    )

    loop.set_category(SHIFT_LOOP_CATEGORY)
    loop.add_tag(tags)

    for shift in shift_list.shifts:
        value_uncertainty = (
            shift.value_uncertainty if shift.value_uncertainty else UNUSED
        )

        element = shift.atom.element if shift.atom.element else UNUSED

        isotope_number = (
            shift.atom.isotope_number if shift.atom.isotope_number else UNUSED
        )

        row_data = {
            "chain_code": shift.atom.residue.chain_code,
            "sequence_code": shift.atom.residue.sequence_code,
            "residue_name": shift.atom.residue.residue_name,
            "atom_name": shift.atom.atom_name,
            "value": shift.value,
            "value_uncertainty": value_uncertainty,
            "element": element,
            "isotope_number": isotope_number,
        }
        loop.add_data(
            [
                row_data,
            ]
        )

    return frame


def collapse_common_shifts(shift_list: ShiftList) -> ShiftList:
    """
    replace entries in the shift list which are in the same residue and are related by symmetry
    and have the same shift by the reduced form e.g. HA1 1.000 HA2 1.000 -> HA* 1.000
    :param shift_list:  shift list to collapse
    :return:  collapsed chemcial shifts
    """

    result = []
    by_residue = _cluster_shifts_by_residue(shift_list)
    for residue_shifts in by_residue.values():
        result.extend(_cluster_by_shift(residue_shifts))

    return ShiftList(result)


def _cluster_shifts_by_residue(
    shift_list: ShiftList,
) -> Dict[SequenceResidue, List[ShiftData]]:
    """
    put all the shifts from the same residue in an entry in a dictionary keyed by residue

    :param shift_list: the ShiftList to cluster
    :return: the dictionary of shifts with the same residue
    """
    by_residue = {}
    for shift in shift_list.shifts:
        atom = shift.atom
        residue = SequenceResidue(
            shift.atom.chain_code, atom.sequence_code, atom.residue_name
        )
        by_residue.setdefault(residue, []).append(shift)

    return by_residue


def _cluster_by_shift(shifts: List[ShiftData]) -> List[List[ShiftData]]:
    """
    build a list lists where each child list contains ShiftData entries
    which all have the same chemicl shifts
    :param shifts: a list of shifts to cluster
    :return: ShiftData clustered by shifts
    """
    shift_map = {}

    for shift in shifts:
        shift_map.setdefault(shift.value, []).append(shift)

    result = []
    for entry in shift_map.values():

        if len(entry) > 1:
            result.extend(_collapse_cluster(entry))
        else:
            result.append(entry[0])

    return result


def _collapse_cluster(shifts: List[ShiftData]) -> List[ShiftData]:
    """
    collapse a cluster of atoms with the same shift which are symmetry related (currently in the nef style)
    HA1 / HA2 -> HA*, HG11, HG12, HG13, HG21, HG22, HG23 -> HG* [not HG**] Note all symmetry related shifts
    should have the same shift else information will be lost

    :param shifts: a list of shifts to cluster
    :return: the clusted shifts as a new list
    """

    by_stem = {}
    for shift in shifts:
        last_character = shift.atom.atom_name[-1]
        if last_character.isnumeric() or last_character == "*":
            by_stem.setdefault(shift.atom.atom_name[:-1], []).append(shift)

    result = []
    for stem, shifts in by_stem.items():
        if len(shifts) > 1:
            first_shift = shifts[0]
            first_atom = first_shift.atom

            residue = SequenceResidue(
                first_atom.chain_code, first_atom.sequence_code, first_atom.residue_name
            )
            new_atom = AtomLabel(residue, f"{stem}*")
            new_shift = ShiftData(
                new_atom, first_shift.value, first_shift.value_uncertainty
            )

            result.append(new_shift)
        else:
            result.append(shifts[0])

    if len(result) != len(shifts):
        result = _collapse_cluster(_collapse_cluster(result))

    return result



import sys
from dataclasses import replace

import pydantic

from nef_pipelines.lib.structures import AtomLabel, SequenceResidue
from nef_pipelines.lib.translation.io import CHEM_COMPS, load_chem_comps
from nef_pipelines.lib.util import exit_error

TRANSLATION_CACHE = {}


def exit_no_naming_system(system):
    msg = f" the atom name translator can't find the naming system {system}"
    exit_error(msg)


def translate_from(
    atom: AtomLabel, system: str, mol_type: str = "protein"
) -> AtomLabel:

    cache_key = atom, mol_type, system

    if cache_key in TRANSLATION_CACHE:
        translated_atom_name = TRANSLATION_CACHE[cache_key]
    else:

        key = mol_type.upper(), atom.residue.residue_name.upper()
        chem_comp = CHEM_COMPS[key]

        sought_naming_system = None
        for naming_system in chem_comp.namingSystems:
            if naming_system.name.upper() == system.upper():
                sought_naming_system = naming_system

        if sought_naming_system is None:
            exit_no_naming_system(system)

        atom_name = atom.atom_name.upper()
        translated_atom_name = atom_name

        ambiguity_mappings = [
            app_data
            for app_data in chem_comp.applicationData
            if app_data.application == "ccpNmr"
            and app_data.keyword == "ambiguityMapping"
        ]
        mapping = [
            ambiguity_mapping.value[1]
            for ambiguity_mapping in ambiguity_mappings
            if ambiguity_mapping.value[0].upper() == atom_name.upper()
        ]

        if len(mapping) == 1:
            translated_atom_name = mapping[0]
        elif len(mapping):
            print(f"unexpected multiple mapping {mapping}", file=sys.stderr)
        else:
            for atom_sys_name in sought_naming_system.atomSysNames:
                if atom_sys_name.sysName.upper() == atom_name.upper():
                    translated_atom_name = atom_sys_name.atomName

        TRANSLATION_CACHE[cache_key] = translated_atom_name

    return (
        replace(atom, atom_name=translated_atom_name)
        if translated_atom_name != atom.atom_name
        else atom
    )


if __name__ == "__main__":

    from codetiming import Timer

    print(pydantic.version.version_info())

    with Timer():
        load_chem_comps()

    atom = AtomLabel(SequenceResidue("A", 1, "VAL"), "HG1##")
    print("start", atom)
    with Timer():
        print("translate", translate_from(atom, "xplor"))
else:
    load_chem_comps()



from pathlib import Path
from typing import List

import typer
from fyeah import f
from pynmrstar import Entry, Loop

from nef_pipelines.lib.nef_lib import (
    UNUSED,
    create_nef_save_frame,
    read_file_or_exit,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.sequence_lib import (
    exit_if_chain_not_in_sequence,
    sequence_from_entry,
)
from nef_pipelines.transcoders.nmrpipe.nmrpipe_lib import read_db_file_records
from nef_pipelines.transcoders.talos import import_app
from nef_pipelines.transcoders.talos.talos_lib import read_talos_secondary_structure

app = typer.Typer()

NEF_PIPELINES_NAMESPACE = "nefpls"
PIPELINES_SECONDARY_STRUCTURE_CATEGORY = (
    f"{NEF_PIPELINES_NAMESPACE}_secondary_structure"
)
DEFAULT_SEPARATOR = "::"
DEFAULT_END = ""


FRAME_NAME_HELP = """\
                        name for the frame that will be created,
                        default is {chain_code}_talos
                    """


@import_app.command()
def secondary_structure(
    chain_code: str = typer.Option(
        "A",
        "--chain",
        help="chain codes for the new chain",
        metavar="<CHAIN-CODE>",
    ),
    frame_name: str = typer.Option(
        "{chain_code}_talos", "-f", "--frame", help=FRAME_NAME_HELP
    ),
    input: Path = typer.Option(
        None,
        "-i",
        "--input",
        metavar="|INPUT|",
        help="file to read NEF data from",
    ),
    include_predictions: bool = typer.Option(False, help="include predictions"),
    file_name: Path = typer.Argument(
        ..., help="input files of type <TALOS-FILE>.pred", metavar="<PRED-FILE>"
    ),
):
    """-  convert talos secondary structure predictions to NEF [namespace nefpls]"""

    entry = read_or_create_entry_exit_error_on_bad_file(input)

    lines = read_file_or_exit(file_name)

    entry = pipe(entry, lines, chain_code, frame_name, file_name, include_predictions)

    print(entry)


def pipe(
    entry: Entry,
    lines: List[str],
    chain_code: str,
    frame_name: str,
    file_name: Path,
    include_predictions: bool,
):

    exit_if_chain_not_in_sequence(chain_code, entry, file_name)

    nef_sequence = sequence_from_entry(entry)

    gdb_records = read_db_file_records(lines)

    talos_secondary_structure = read_talos_secondary_structure(
        gdb_records,
        chain_code=chain_code,
        file_name=file_name,
        nef_sequence=nef_sequence,
        include_predictions=include_predictions,
    )

    frame_name = f(frame_name)

    frame = create_nef_save_frame(PIPELINES_SECONDARY_STRUCTURE_CATEGORY, frame_name)

    extra_tags = {"method": "talos", "version": UNUSED}

    for tag, value in extra_tags.items():
        frame.add_tag(tag, value)

    data_loop = Loop.from_scratch(f"{NEF_PIPELINES_NAMESPACE}_secondary_structure")

    tags = """
        index
        chain_code sequence_code residue_name
        secondary_structure merit comment
        """.lower().split()

    for tag in tags:
        data_loop.add_tag(tag)

    frame.add_loop(data_loop)

    for index, secondary_structure in enumerate(talos_secondary_structure, start=1):

        residue = secondary_structure.residue
        row = {
            "index": index,
            "chain_code": residue.chain_code,
            "sequence_code": residue.sequence_code,
            "residue_name": residue.residue_name,
            "secondary_structure": secondary_structure.secondary_structure,
            "merit": secondary_structure.merit,
            "comment": secondary_structure.comment,  # TODO: lose whole column if there are none...
        }

        data_loop.add_data(
            [
                row,
            ]
        )

    entry.add_saveframe(frame)

    return entry



import string

import pytest

from nef_pipelines.lib.util import strip_characters_left, strip_characters_right


@pytest.mark.parametrize(
    "input, expected",
    [
        ("", ("", "")),
        ("AAAA", ("", "AAAA")),
        ("123AAAA", ("123", "AAAA")),
    ],
)
def test_strip_characters_left(input, expected):

    result = strip_characters_left(input, string.digits)

    assert result == expected


@pytest.mark.parametrize(
    "input, expected",
    [
        ("", ("", "")),
        ("AAAA", ("AAAA", "")),
        ("AAAA123", ("AAAA", "123")),
    ],
)
def test_strip_characters_right(input, expected):

    result = strip_characters_right(input, string.digits)
    assert result == expected



import string
from dataclasses import replace
from functools import partial
from pathlib import Path
from typing import List

import typer
from fyeah import f

from nef_pipelines.lib.isotope_lib import Isotope
from nef_pipelines.lib.nef_lib import (
    UNUSED,
    add_frames_to_entry,
    read_or_create_entry_exit_error_on_bad_file,
)
from nef_pipelines.lib.peak_lib import peaks_to_frame
from nef_pipelines.lib.sequence_lib import (
    MoleculeTypes,
    get_residue_name_from_lookup,
    sequence_from_entry_or_exit,
    sequence_to_residue_name_lookup,
)
from nef_pipelines.lib.translation_lib import translate_new_peak
from nef_pipelines.lib.util import STDIN, exit_error
from nef_pipelines.transcoders.mars import import_app
from nef_pipelines.transcoders.sparky.importers.peaks import (
    _guess_dimensions_if_not_defined_or_throw,
)
from nef_pipelines.transcoders.sparky.sparky_lib import (
    parse_peaks as parse_sparky_peaks,
)

MAKE_SPECTA_HELP = """
    if true merge relevant files to give spectra frames [e.g combine Ca and Ca-1 to get an HNCA frame and select Ca-1 to
    get an HNcoCA frame...] files with CA/CB/CB and CA-1/CB-1/CO-1 in their names are combined, spectra without these
    selectors in their names are read as individual spectra [not filenames will be replaced by spectrum names in the
    saveframe template]
"""


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def peaks(
    chain_code: str = typer.Option(
        "A",
        "--chain",
        help="chain code to use, as mars can only assign a single chain a single chaincode is assumed here",
        metavar="<CHAIN-CODE>",
    ),
    frame_name: str = typer.Option(
        "mars_{file_name}",
        "-f",
        "--frame-name",
        help="a templated name for the frame {file_name} will be replaced by input filename",
    ),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        metavar="|PIPE|",
        help="input to read NEF data from [- is stdin]",
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="input files of type peaks.txt", metavar="<MARS-peaks>.txt"
    ),
    spectrometer_frequency: float = typer.Option(
        600.123456789, help="spectrometer frequency in MHz"
    ),
    include_unassigned: bool = typer.Option(False, help="include unassigned peaks"),
    make_spectra: bool = typer.Option(True, help=MAKE_SPECTA_HELP),
    dont_sort_peaks: bool = typer.Option(
        False, help="don't sort the peaks [by isotopes 1H and then 13C]"
    ),
):
    """- import MARS [sparkyish] peaks files sparky_<NAME>.out"""

    entry = read_or_create_entry_exit_error_on_bad_file(input)

    sequence = sequence_from_entry_or_exit(entry)

    file_names_and_lines = {}
    for file_name in file_names:

        try:
            with open(file_name, "r") as fp:
                lines = fp.readlines()
        except IOError as e:
            msg = f"""
                    while reading sparky peaks file {file_name} there was an error reading the file
                    the error was: {e}
                """
            exit_error(msg, e)

        file_names_and_lines[file_name] = lines

    entry = pipe(
        entry,
        frame_name,
        file_names_and_lines,
        chain_code,
        sequence,
        spectrometer_frequency=spectrometer_frequency,
        include_unassigned=include_unassigned,
        make_spectra=make_spectra,
        sort_peaks=not dont_sort_peaks,
    )

    print(entry)


def pipe(
    entry,
    frame_code_template,
    file_names_and_lines,
    chain_code,
    sequence,
    spectrometer_frequency,
    include_unassigned=False,
    make_spectra=True,
    sort_peaks=True,
):
    delete_unassigned = not include_unassigned

    if make_spectra:
        spectra_sets = _combine_filenames_into_spectra(file_names_and_lines.keys())
    else:
        spectra_sets = {
            file_name.stem: [
                file_name,
            ]
            for file_name in file_names_and_lines.keys()
        }

    peaks_by_filename = {}
    for file_name, lines in file_names_and_lines.items():
        peaks = _parse_peaks(chain_code, delete_unassigned, file_name, lines, sequence)
        peaks = [translate_new_peak(peak) for peak in peaks]
        peaks_by_filename[file_name] = peaks

    peaks_by_frame_name = {}
    for frame_name, file_names in spectra_sets.items():
        for file_name in file_names:
            peaks_by_frame_name.setdefault(frame_name, []).extend(
                peaks_by_filename[file_name]
            )

    mars_frames = []

    for frame_name, peaks in peaks_by_frame_name.items():
        peaks = list(peaks)

        # peaks.sort() # need to sort peaks on amides...

        dimensions = _guess_dimensions_if_not_defined_or_throw(peaks, [])

        dimensions = [{"axis_code": dimension} for dimension in dimensions]

        if sort_peaks:
            sort_isotopes = [Isotope.H1, Isotope.C13]
            peaks = _sort_by_isotopes(peaks, dimensions, sort_isotopes)

        file_name = frame_name  # used in f method...

        frame_code = f(frame_code_template)

        frame = peaks_to_frame(
            peaks, dimensions, spectrometer_frequency, frame_code=frame_code
        )

        mars_frames.append(frame)

    return add_frames_to_entry(entry, mars_frames)


def _sort_by_isotopes(peaks, dimensions, sort_isotopes):
    axis_dimensions = {axis["axis_code"]: i for i, axis in enumerate(dimensions)}
    sort_dimensions = [axis_dimensions[isotope] for isotope in sort_isotopes]

    def sort_function(value, dimensions):
        result = []
        for dimension in dimensions:
            residue = value.shifts[dimension].atom.residue
            result.append(residue.chain_code)
            result.append(int(residue.sequence_code))
        return result

    sort_curry = partial(sort_function, dimensions=sort_dimensions)

    return sorted(peaks, key=sort_curry)


def _parse_peaks(chain_code, delete_unassigned, file_name, lines, sequence):

    lookup = sequence_to_residue_name_lookup(sequence)

    new_lines = []
    for line_no, line in enumerate(lines, start=1):
        line = line.strip()

        if len(line) == 0:
            continue

        fields = line.split()

        if line_no == 1:
            number_fields = len(fields)
            shifts = " ".join([f"w{i}" for i in range(1, number_fields)])
            new_lines.append(f"Assignment {shifts}")

        is_assignment = False
        for quality in "LMH":
            if f"({quality})" in line:
                is_assignment = True

        assignment = fields[0].replace("(", "|").replace(")", "|")
        assignment = assignment.strip("|")
        assignment_fields = assignment.split("|")

        if is_assignment:
            fields[0] = f"{assignment_fields[0]}{assignment_fields[1]}"
        else:
            pseudo_number = "".join(
                [char for char in assignment_fields[0] if char.isdigit()]
            )
            if len(pseudo_number) == 0:
                pseudo_number = hash(assignment_fields[0])
            fields.append(f"PSEUDO_RESIDUE={assignment_fields[0]}")
            fields[0] = f"PR{pseudo_number}{assignment_fields[1]}"

        if is_assignment:
            fields.append(f"merit={assignment_fields[-1]}")

        new_line = " ".join(fields)

        new_lines.append(new_line)

    peaks = parse_sparky_peaks(
        new_lines,
        file_name=file_name,
        molecule_type=MoleculeTypes.PROTEIN,
        chain_code=chain_code,
        sequence=sequence,
        allow_pseudo_atoms=True,
    )
    modified_peaks = []
    for peak in peaks:
        if "PSEUDO_RESIDUE" in peak.comment:

            if delete_unassigned:
                continue

            pseudo_residue = peak.comment.split("=")[1]
            peak = replace(peak, comment="")
            new_shifts = []
            for shift in peak.shifts:
                residue = shift.atom.residue
                # is this the best thing to do?
                residue = replace(
                    residue,
                    chain_code="@-",
                    residue_name=UNUSED,
                    sequence_code=f"@{pseudo_residue}",
                )
                atom = replace(shift.atom, residue=residue)
                shift = replace(shift, atom=atom)
                new_shifts.append(shift)
            peak = replace(peak, shifts=new_shifts)

        modified_peaks.append(peak)
    peaks = modified_peaks
    modified_peaks = []
    for peak in peaks:
        new_shifts = []
        for shift in peak.shifts:
            if shift.atom.atom_name[-1] in string.ascii_lowercase:

                residue = shift.atom.residue
                if residue.residue_name == UNUSED:
                    residue = replace(
                        residue, sequence_code=f"{residue.sequence_code}-1"
                    )
                else:
                    offset_sequence_code = int(residue.sequence_code) - 1
                    chain_code = residue.chain_code

                    offset_residue_name = get_residue_name_from_lookup(
                        chain_code, offset_sequence_code, lookup
                    )
                    residue = replace(
                        residue,
                        sequence_code=f"{offset_sequence_code}",
                        residue_name=offset_residue_name,
                    )

                atom = replace(shift.atom, atom_name=shift.atom.atom_name.upper())
                atom = replace(atom, residue=residue)
                shift = replace(shift, atom=atom)
            new_shifts.append(shift)

        modified_peaks.append(replace(peak, shifts=new_shifts))
    peaks = modified_peaks
    modified_peaks = []
    MERITS = {"L": 0.50, "M": 0.75, "H": 1.00}
    for peak in peaks:
        if "merit" in peak.comment:
            merit_letter = peak.comment[-1]

            peak = replace(peak, figure_of_merit=MERITS[merit_letter])
        modified_peaks.append(peak)
    peaks = modified_peaks
    return peaks


def _combine_filenames_into_spectra(file_names):

    used_filenames = set()

    spectra_and_selectors = {
        "HNCA": ["_CA.", "_CA-1."],
        "HNcoCA": ["_CA-1."],
        "HNCACB": ["_CA.", "_CA-1.", "_CB.", "_CB-1."],
        "HNcoCACB": ["_CA-1.", "_CB-1."],
        "HNCO": ["_CO-1."],
        "HNcaCO": ["_CO.", "_CO-1."],
    }

    combinations = {}

    for spectrum_name, selectors in spectra_and_selectors.items():
        selected = []
        for selector in selectors:
            for file_name in file_names:
                if selector in file_name.name:
                    selected.append(file_name)
                    used_filenames.add(file_name)
                    break

        if len(selected) == len(selectors):
            combinations[spectrum_name] = selected

    unused_filenames = set(file_names) - used_filenames
    for unused_filename in unused_filenames:
        combinations[unused_filename.stem] = [
            unused_filename,
        ]

    return combinations



from pathlib import Path
from typing import List

import typer
from pynmrstar import Entry

from nef_pipelines.lib.nef_lib import read_or_create_entry_exit_error_on_bad_file
from nef_pipelines.lib.sequence_lib import sequence_to_nef_frame
from nef_pipelines.lib.structures import SequenceResidue
from nef_pipelines.lib.util import STDIN, parse_comma_separated_options
from nef_pipelines.transcoders.xeasy import import_app
from nef_pipelines.transcoders.xeasy.xeasy_lib import parse_sequence

app = typer.Typer()

NO_CHAIN_START_HELP = """don't include the start chain link type on a chain for the first residue [linkage will be
                            middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call
                            this option multiple times to set chain starts for multiple chains"""
NO_CHAIN_END_HELP = """don't include the end chain link type on a chain for the last residue [linkage will be
                        middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call this
                        option multiple times to set chain ends for multiple chains"""


@import_app.command()
def sequence(
    entry_name: str = typer.Option("xeasy", help="a name for the entry if required"),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        metavar="|PIPE|",
        help="pipe to read NEF data from, for testing [overrides stdin !use stdin instead!]",
    ),
    no_chain_starts: List[str] = typer.Option(
        [], "--no-chain-start", help=NO_CHAIN_START_HELP
    ),
    no_chain_ends: List[str] = typer.Option(
        [], "--no-chain-end", help=NO_CHAIN_END_HELP
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="the file to read", metavar="<XEASY-SEQUENCE-FILE.seq>"
    ),
):
    """- convert xeasy sequences to nef"""

    no_chain_starts = parse_comma_separated_options(no_chain_starts)
    if not no_chain_starts:
        no_chain_starts = [False]

    no_chain_ends = parse_comma_separated_options(no_chain_ends)
    if no_chain_ends:
        no_chain_ends = [False]

    entry = read_or_create_entry_exit_error_on_bad_file(input, entry_name)

    entry = pipe(entry, no_chain_starts, no_chain_ends, file_names)

    print(entry)


def pipe(
    entry: Entry,
    no_chain_starts: List[bool],
    no_chain_ends: List[bool],
    file_names: List[Path],
):

    xeasy_sequences = set()

    for file_name in file_names:

        xeasy_sequences.update(_read_sequence(file_name))

    xeasy_sequences = sorted(xeasy_sequences)

    xeasy_frame = sequence_to_nef_frame(xeasy_sequences, no_chain_starts, no_chain_ends)

    entry.add_saveframe(xeasy_frame)

    return entry


def _read_sequence(file_name: Path) -> List[SequenceResidue]:
    return parse_sequence(file_name.open().readlines())



import typer

from nef_pipelines import nef_app

app = typer.Typer()
import_app = typer.Typer()
export_app = typer.Typer()

if nef_app.app:

    nef_app.app.add_typer(
        app,
        name="nmrview",
        help="-  read and write nmrview [peaks, sequences & shifts]",
    )

    app.add_typer(
        import_app, name="import", help="-  import nmrview [peaks, sequences & shifts]"
    )
    app.add_typer(
        export_app, name="export", help="-  export nmrview [peaks, sequences & shifts]"
    )

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.transcoders.nmrview.exporters.peaks  # noqa: F401
    import nef_pipelines.transcoders.nmrview.exporters.sequences  # noqa: F401
    import nef_pipelines.transcoders.nmrview.exporters.shifts  # noqa: F401
    import nef_pipelines.transcoders.nmrview.importers.peaks  # noqa: F401
    import nef_pipelines.transcoders.nmrview.importers.sequence  # noqa: F401
    import nef_pipelines.transcoders.nmrview.importers.shifts  # noqa: F401



# # todo sort out app data lists
# # todo derived methods
# # todo remove parent links on json dump
# # todo unify tree iterators!
#

from enum import Enum, auto
from typing import (
    Annotated,
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    TypeVar,
    Union,
    get_args,
    get_origin,
)

from annotated_types import Gt

#
from cachetools import LRUCache
from pydantic import BaseModel, Field, GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema
from pydantic_core.core_schema import no_info_after_validator_function
from strenum import StrEnum

from nef_pipelines.lib.translation.object_iter import (
    ObjectIter,
    item_only,
    object_children,
)


class PCMeta(BaseModel.__class__):
    _object_map = {}

    def __call__(self, *args, **kwargs):
        result = super(PCMeta, self).__call__(*args, **kwargs)

        if hasattr(result, "__post_init__") and isinstance(
            result.__post_init__, Callable
        ):
            result.__post_init__()

        return result


class ID(int):
    @classmethod
    def __get_pydantic_core_schema__(
        cls, _: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(
            cls, handler(Annotated[int, Gt(0)])
        )


T = TypeVar("T")


class IDTuple(Tuple):
    is_idlist = True

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: Callable[[Any], CoreSchema]
    ) -> CoreSchema:
        args = get_args(source_type)
        if args:
            return handler(Tuple[args])
        else:
            return no_info_after_validator_function(IDTuple, handler(Tuple))


class IDList(List):
    is_idlist = True

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: Callable[[Any], CoreSchema]
    ) -> CoreSchema:
        args = get_args(source_type)
        if args:
            return handler(List[args])
        else:
            return no_info_after_validator_function(IDList, handler(List))


# these are distinguished by containing IDLists / IDTuples
# these trigger replacement of IDs by instances
IDs = Union[IDList[ID], ID]
Types = Union[List[T], T]
OptionalIDs = Optional[IDs]

# containers of a specific type [child containers] - only some of these are created on their own as indicated
ItemOfType = Annotated[T, None]
PairOfType = Tuple[T, T]  # not created standalone
TripletOfType = Tuple[T, T, T]  # not created standalone
QuartetOfType = Tuple[T, T, T, T]  # not created standalone
OneOrMoreOfType = Union[List[T], T]
OptionalOneOrMoreOfType = Optional[OneOrMoreOfType]

# containers for IDs - not used on their own
ItemOfID = Annotated[ID, None]
PairOfID = IDTuple[ID, ID]
TripletOfID = IDTuple[ID, ID, ID]
QuartetOfID = IDTuple[ID, ID, ID, ID]
OneOrMoreOfID = Union[IDList[ID], ID]
OptionalOneOrMoreOfID = Optional[OneOrMoreOfID]  # - no such instances currently

# containers of IDs or types [links]
Item = Union[T, ID]
Pair = Union[PairOfType[T], PairOfID]
Triplet = Union[TripletOfType[T], TripletOfID]
Quartet = Union[QuartetOfType[T], QuartetOfID]
OneOrMore = Union[OneOrMoreOfType[T], OneOrMoreOfID]

OptionalItem = Optional[Item[T]]
OptionalOneOrMore = Optional[OneOrMore[T]]


class Linkage(StrEnum):
    parent = auto()
    children = auto()
    link = auto()
    attribute = auto()
    implementation = auto()
    derived_link = auto()
    derived_attribute = auto()
    cross_package_link = auto()

    none = auto()


# # attribute that is a parent
Parent = Annotated[
    Optional[ItemOfType[T]], Linkage.parent
]  # optional as they are set at runtime

# attribute that is a container of a child or children
Children = Annotated[T, Linkage.children]

# attribute that is a link between objects but not a child
Link = Annotated[T, Linkage.link]

WithInverse = Annotated

Attribute = Annotated[T, Linkage.attribute]

Implementation = Annotated[T, Linkage.implementation]

DerivedLink = Annotated[
    Optional[T], Linkage.derived_link
]  # optional as they are set at runtime

CrossPackageLink = Annotated[
    Optional[T], Linkage.cross_package_link
]  # optional as they are set at runtime


class PCBaseModel(BaseModel, metaclass=PCMeta):

    ID: Implementation[Optional[int]] = (
        None  # not required by appdata maybe we need further level of base model
    )

    @classmethod
    def _get_field_annotations(cls, field_name):
        annotations = (
            cls.__annotations__[field_name] if field_name in cls.__annotations__ else []
        )
        if annotations and not isinstance(annotations, list):
            annotations = [annotations]

        return annotations

    @classmethod
    def _field_is_id_list(cls, field_name):
        annotations = cls._get_field_annotations(field_name)

        origin = get_origin(annotations)
        result = hasattr(origin, "is_idlist")

        if not result:
            args = get_args(annotations)
            arg_origins = [
                hasattr(get_origin(arg), "is_idlist") for arg in args if get_origin(arg)
            ]
            result = True in arg_origins

        return result

    def __init__(self, **kwargs):
        super(PCBaseModel, self).__init__(**kwargs)

        if self.ID is not None:
            # noinspection PyProtectedMember,PyUnresolvedReferences
            self.__class__._object_map[self.ID] = self

    def __str__(self) -> str:
        return f"<class {str(self.__class__)[8:-2]} {self.ID}>"

    __repr__ = __str__


# forward references for type system
class ChemAtom(PCBaseModel): ...


class ChemBond(PCBaseModel): ...


class ChemCompVar(PCBaseModel): ...


class ChemTorsion(PCBaseModel): ...


class LinkEnd(PCBaseModel): ...


class ChemTorsionSysName(PCBaseModel): ...


class ChemAtomSet(PCBaseModel): ...


class ChemCompSysName(PCBaseModel): ...


class NamingSystem(PCBaseModel): ...


class ChemAngle(PCBaseModel): ...


class AbstractChemAtom(PCBaseModel): ...


class LinkAtom(PCBaseModel): ...


class ChemComp(PCBaseModel): ...


class AppDataString(PCBaseModel): ...


class AppDataInt(PCBaseModel): ...


class AppDataBoolean(PCBaseModel): ...


class InverseName(str): ...


def name_annotation(name):
    return InverseName(name)


class AtomChirality(StrEnum):
    R = auto()
    S = auto()
    unknown = auto()


class Stereochemistry(PCBaseModel):
    serial: Attribute[int]  # 1..1
    stereoClass: Attribute[str]  # 1..1
    value: Attribute[str]  # 1..1

    chemAtoms: Link[
        WithInverse[
            OptionalOneOrMore[AbstractChemAtom], name_annotation("stereochemstries")
        ]
    ]  # 0..*
    chemComp: Parent[ChemComp]  # 1..1
    coreAtoms: Link[
        WithInverse[
            OptionalOneOrMore[AbstractChemAtom],
            name_annotation("coreStereochemistries"),
        ]
    ]  # 0..*
    parent: Parent[ChemComp]  # 1..1
    # package             # 0..1


class ChemAngle(PCBaseModel):
    chemAtoms: Link[
        WithInverse[Triplet[ChemAtom], name_annotation("chemAngles")]
    ]  # 3..3
    chemComp: Parent[ChemComp]  # 1..1
    parent: Parent[ChemComp]  # 1..1


class AbstractChemAtom(PCBaseModel):
    name: Attribute[str]  # 1..1
    subType: Attribute[int] = 1  # 1..1

    chemAngles: Link[
        WithInverse[OptionalOneOrMore[ChemAngle], name_annotation("chemAtoms")]
    ]  # 0..*
    chemBonds: Link[
        WithInverse[OptionalOneOrMore[ChemBond], name_annotation("chemAtoms")]
    ]  # 0..*
    chemComp: Parent[ChemComp]  # 1..1
    chemCompVars: Link[
        WithInverse[OptionalOneOrMore[ChemCompVar], name_annotation("chemAtoms")]
    ]  # 0..*
    chemTorsions: Link[
        WithInverse[OptionalOneOrMore[ChemTorsion], name_annotation("chemAtoms")]
    ]  # 0..*
    coreStereochemistries: Link[
        WithInverse[OptionalOneOrMore[Stereochemistry], name_annotation("coreAtoms")]
    ]  # 0..*
    parent: Parent[ChemComp]  # 1..1
    stereochemistries: Link[
        WithInverse[OptionalOneOrMore[Stereochemistry], name_annotation("chemAtoms")]
    ]  # 0..*


class ChemAtom(AbstractChemAtom):
    type: Attribute[Literal["ChemAtom"]]
    chirality: Attribute[OptionalItem[AtomChirality]]  # 0..1
    elementSymbol: Attribute[str]  # 1..1
    nuclGroupType: Attribute[Optional[str]]  # 0..1
    shortVegaType: Attribute[Optional[str]]  # 0..1
    waterExchangeable: Attribute[bool] = False  # 1..1

    boundLinkEnds: Link[
        WithInverse[OptionalOneOrMore[LinkEnd], name_annotation("boundChemAtom")]
    ]  # 0..*
    chemAtomSet: Link[
        WithInverse[OptionalItem[ChemAtomSet], name_annotation("chemAtoms")]
    ]  # 0..1
    chemComp: Parent[ChemComp]  # 1..1
    # chemElement # derived                               # 1..1
    parent: Parent[ChemComp]  # 1..1
    remoteLinkEnds: Link[
        WithInverse[OptionalOneOrMore[LinkEnd], name_annotation("remoteChemAtom")]
    ]  # 0..*


class LinkAtom(AbstractChemAtom):
    type: Literal["LinkAtom"]
    elementSymbol: Optional[str] = None  # 0..1

    boundLinkEnd: Link[
        WithInverse[OptionalItem[LinkEnd], name_annotation("boundLinkAtom")]
    ]  # 0..1
    chemComp: Parent[ChemComp]  # 1..1
    parent: Parent[ChemComp]  # 1..1
    remoteLinkEnd: Link[
        WithInverse[OptionalItem[LinkEnd], name_annotation("remoteLinkAtom")]
    ]  # 0..1


class ChemTorsion(PCBaseModel):
    name: Attribute[str]  # 0..1

    chemAtoms: Link[
        WithInverse[Quartet[ChemAtom], name_annotation("chemTorsions")]
    ]  # 4..4
    chemComp: Parent[ChemComp]  # 1..1
    parent: Parent[ChemComp]  # 1..1
    sysNames: Link[
        WithInverse[
            OptionalOneOrMore[ChemTorsionSysName], name_annotation("chemTorsion")
        ]
    ]  # 0..*


class ChemAtomSet(PCBaseModel):
    distCorr: Attribute[float] = 0.0  # 1..1
    # elementSymbol  #  derived                            # 1..1
    isEquivalent: Attribute[Optional[bool]]  # 0..1
    isProchiral: Attribute[
        Optional[bool]
    ]  # 1..1  # no default value but not always present
    name: Attribute[str]  # 1..1
    subType: Attribute[int] = 1  # 1..1

    chemAtomSet: Link[
        WithInverse[OptionalItem[ChemAtomSet], name_annotation("chemAtomSets")]
    ]  # 0..1
    chemAtomSets: Link[
        WithInverse[OptionalOneOrMore[ChemAtomSet], name_annotation("chemAtomSet")]
    ]  # 0..*
    chemAtoms: Link[
        WithInverse[OptionalOneOrMore[ChemAtom], name_annotation("chemAtomSet")]
    ]  # 1..1
    chemComp: Parent[ChemComp]  # 1..1
    parent: Parent[ChemComp]  # 1..1


class BondType(StrEnum):
    single = auto()
    double = auto()
    triple = auto()
    aromatic = auto()
    dative = auto()
    singleplanar = auto()


class BondStereochemistry(StrEnum):
    E = auto()
    Z = auto()
    unknown = auto()


class ChemBond(PCBaseModel):
    bondType: Attribute[BondType] = BondType.single  # 1..1
    stereochem: Attribute[Optional[BondStereochemistry]]  # 0..1

    chemAtoms: Link[WithInverse[Pair[ChemAtom], name_annotation("chemBonds")]]  # 2..2
    chemComp: Parent[ChemComp]  # 1..1
    parent: Parent[ChemComp]  # 1..1


class ChemCompLinking(StrEnum):
    start = auto()
    middle = auto()
    end = auto()
    free = auto()
    none = auto()
    any = auto()


class ChemCompSysName(PCBaseModel):
    sysName: Attribute[str]  # 1..1

    # chemCompVars # derived                                # 0..*
    namingSystem: Parent[NamingSystem]  # 1..1
    # TODO this name is incorrect...
    specificChemCompvars: Link[
        WithInverse[OptionalOneOrMore[ChemCompVar], name_annotation("specificSysNames")]
    ]  # 0..*
    parent: Parent[NamingSystem]  # 1..1


class ChemCompVar(PCBaseModel):
    descriptor: Attribute[str]  # 1..1
    formalCharge: Attribute[Optional[int]]  # 1..1 but apparently no default...
    # formula # derived str
    glycoCtCode: Attribute[Optional[str]]  # 0..1
    isAromatic: Attribute[bool] = False  # 1..1 but apparently no default...
    isDefaultVar: Attribute[bool] = False  # 1..1
    isParamagnetic: Attribute[bool] = False  # 1..1 but apparently no default...
    linking: Attribute[ChemCompLinking]  # 1..1
    # molecularMass derived                                   # 1..1
    # name derived                                            # 0..1
    nonStereoSmiles: Attribute[Optional[str]]  # 0..1
    stereoSmiles: Attribute[Optional[str]]  # 0..1
    varName: Attribute[Optional[str]]  # 0..1

    # chemAngles    derived                                   # 0..*
    # chemAtomSets  derived                                   # 0..*
    chemAtoms: Link[
        WithInverse[
            OptionalOneOrMore[AbstractChemAtom], name_annotation("chemCompVars")
        ]
    ]  # 0..*
    # chemBonds     derived                                   # 0..*
    chemComp: Parent[ChemComp]  # 1..1
    # chemCompSysNames derived                                # 0..*
    # chemTorsions derived                                    # 0..*
    # linkEnds derived                                        # 0..*
    parent: Parent[ChemComp]  # 1..1
    specificSysNames: Link[
        WithInverse[
            OptionalOneOrMore[ChemCompSysName], name_annotation("specificChemCompVars")
        ]
    ]  # 0..*


class LinkEnd(PCBaseModel):
    linkCode: Attribute[str]  # should be enum                         # 1..1

    boundChemAtom: Link[
        WithInverse[OptionalItem[ChemAtom], name_annotation("boundLinkEnds")]
    ]  # 1..1 is required but not always present?
    boundLinkAtom: Link[
        WithInverse[OptionalItem[LinkAtom], name_annotation("boundLinkEnd")]
    ]  # 1..1 is required but not always present?
    chemComp: Parent[ChemComp]  # 1..1
    # chemCompVars derived
    parent: Parent[ChemComp]  # 1..1
    remoteChemAtom: Link[
        WithInverse[OptionalItem[ChemAtom], name_annotation("remoteLinkEnds")]
    ]  # 1..1 is required but not always present?
    remoteLinkAtom: Link[
        WithInverse[OptionalItem[LinkAtom], name_annotation("remoteLinkEnd")]
    ]  # 1..1 is required but not always present?


class AtomSysName(PCBaseModel):
    altSysName: Attribute[OptionalOneOrMore[str]]  # 0..*
    atomName: Attribute[str]  # 1..1
    atomSubType: Attribute[int] = 1  # 1..1
    sysName: Attribute[Optional[str]]  # 1..1 is required but not always present?

    namingSystem: Parent[NamingSystem]  # 1..1
    parent: Parent[NamingSystem]  # 1..1


class ChemTorsionSysName(PCBaseModel):
    sysName: Attribute[Optional[str]]  # 1..1
    chemTorsion: Link[
        WithInverse[OptionalItem[ChemTorsion], name_annotation("sysNames")]
    ]  # 1..1 is required but not always present?
    namingSystem: Parent[NamingSystem]  # 1..1
    parent: Parent[NamingSystem]  # 1..1


class NamingSystem(PCBaseModel):
    name: Attribute[str]  # 1..1

    atomReference: Link[
        WithInverse[OptionalItem[NamingSystem], name_annotation("atomVariantSystems")]
    ]  # 0..1
    atomSetReference: Link[
        WithInverse[
            OptionalItem[NamingSystem], name_annotation("atomSetVariantSystems")
        ]
    ]  # 0..1
    atomSetVariantSystems: Link[
        WithInverse[
            OptionalOneOrMore[NamingSystem], name_annotation("atomSetReference")
        ]
    ]  # 0..*
    atomSysNames: Children[OptionalOneOrMore[AtomSysName]]  # 0..*
    atomVariantSystems: Link[
        WithInverse[OptionalOneOrMore[NamingSystem], name_annotation("atomReference")]
    ]  # 0..*
    chemComp: Parent[ChemComp]  # 1..1
    chemCompSysNames: Children[OptionalOneOrMore[ChemCompSysName]]  # 0..*
    chemTorsionSysNames: Children[OptionalOneOrMore[ChemTorsionSysName]]  # 0..*
    # mainChemCompSysName # derived                            # 0..1
    parent: Parent[ChemComp]  # 1..1


class ApplicationData(PCBaseModel):
    application: Attribute[str]  # 1..1
    keyword: Attribute[str]  # 1..1

    def __str__(self):
        return f"{self.__class__} {self.application}:{self.keyword}"

    __repr__ = __str__


class AppDataStr(ApplicationData):
    type: Literal["AppDataString"]
    value: Attribute[
        OneOrMore[str]
    ]  # 1..1 we are getting lists of strings here where it should be lists of AppDataString...


class AppDataInt(ApplicationData):
    type: Literal["AppDataInt"]
    value: Attribute[
        OneOrMore[int]
    ]  # 1..1 we are getting lists of ints here where it should be lists of AppDataString...


class AppDataBoolean(ApplicationData):
    type: Literal["AppDataBoolean"]
    value: Attribute[bool]  # 1..1 seems to be correct here...


AppDataTypes = Union[AppDataStr, AppDataInt, AppDataBoolean]


def object_children_root_only(target, parent):

    linkage = get_linkage(parent.target, parent.selector)
    elems = object_children(target, parent)

    if isinstance(target, (list, PCBaseModel)):
        if linkage in (Linkage.link, Linkage.parent):
            for elem in elems:
                elem.is_leaf = True

    return elems


class MolType(StrEnum):
    protein = auto()
    DNA = auto()
    RNA = auto()
    carbohydrate = auto()
    other = auto()


# inspect.get_annotations only available in 3.10
class AnnotationsCache:
    def __init__(self, size):
        self._cache = LRUCache(size)

    def get_annotations(self, o):

        # object_id = (o.__class__.__name__, o.__class__.__module__, id(o))

        # if object_id in self._cache:
        #      result = self._cache[object_id]
        #
        # else:
        if isinstance(o, type):
            attr_dict = getattr(o, "__dict__", {})
            result = attr_dict.get("__annotations__", {})
        else:
            result = getattr(o, "__annotations__", {})

        if result is None:
            result = {}

        # self._cache[object_id] = result

        return result


# inspect.get_annotations recreates the data everytime so cache it !!!
annotation_cache = AnnotationsCache(3000)


def get_annotations(o: object, include_super_classes: bool = True) -> Dict[str, type]:

    annotations = annotation_cache.get_annotations(o)

    if include_super_classes:
        for super_class in o.__class__.mro():
            super_class_annotations = annotation_cache.get_annotations(super_class)

            super_class_annotations.update(annotations)
            annotations = super_class_annotations

    return annotations


def get_item_linkage(item: ObjectIter.Entry) -> Linkage:
    return get_linkage(item.target, item.selector)


def get_linkage(target: object, selector: str) -> Linkage:

    annotations = get_annotations(target)

    result = Linkage.none
    if selector and selector in annotations:
        annotation = annotations[selector]
        if get_origin(annotation) is Annotated:
            args = set(get_args(annotation))

            link_type = args.intersection(Linkage.__members__.keys())
            result = next(iter(link_type)) if len(link_type) else Linkage.none

    return result


def get_item_opposite(item: ObjectIter.Entry) -> Tuple[str, List[type]]:
    return get_opposite_and_types(item.target, item.selector)


def get_item_opposite_selector(item: ObjectIter.Entry) -> Optional[str]:
    return get_opposite_selector(item.target, item.selector)


def get_opposite_selector(target: object, selector: str) -> Optional[str]:
    annotations = get_annotations(target)

    result = None
    if selector and selector in annotations:
        annotation = annotations[selector]
        if get_origin(annotation) is Annotated:
            args = set(get_args(annotation))

            result = [arg for arg in args if type(arg) == InverseName]

            if len(result) > 1:
                raise Exception(f"unexpected multiple opposites {result}")

    return result[0] if len(result) == 1 else None


def get_opposite_and_types(target: object, selector: str) -> Tuple[str, List[type]]:

    annotations = get_annotations(target)

    inverse = None
    opposite_types = []
    if selector and selector in annotations:
        annotation = annotations[selector]
        if get_origin(annotation) is Annotated:
            args = set(get_args(annotation))

            inverse = [arg for arg in args if type(arg) == InverseName]
            opposite_type = [arg for arg in args if type(arg) != InverseName]

            inverse = inverse[0] if len(inverse) > 0 else None

            if Linkage.link in opposite_type:
                opposite_type.remove(Linkage.link)
                opposite_type = opposite_type[0]

            if get_origin(opposite_type) == Union:
                opposite_type = get_args(opposite_type)

            opposite_types = [
                get_origin(arg) if get_origin(arg) else arg for arg in opposite_type
            ]

    return inverse, opposite_types


def id_in_iter(iterable, object_id):
    result = False
    for value in iterable:
        if isinstance(value, ID) and value == object_id:
            result = True
            break
    return result


def origin_or_value(annotations):
    return [
        get_origin(annotation) if get_origin(annotation) is not None else annotation
        for annotation in annotations
    ]


def strip_annotation(annotations):
    current_annotations = annotations

    do_strip = True
    while do_strip:
        if get_origin(current_annotations) == Annotated:
            current_annotations = get_args(current_annotations)[0]
            continue
        if get_origin(current_annotations) == Union:
            current_annotations = get_args(current_annotations)
            do_strip = False
            continue

        do_strip = False
        continue

    annotations_are_iterable = isinstance(current_annotations, (list, tuple))
    result = (
        origin_or_value(current_annotations)
        if annotations_are_iterable
        else [get_origin(current_annotations)]
    )

    return result


class ContainerTypes(Enum):
    SINGLETON = auto()
    ORDERED = auto()  # a tuple
    UNORDERED = auto()  # a list
    unordered_unique = auto()  # a set - but we currently use a list
    ordered_unique = auto()  # an ordered set - but we currently use a list


def get_link_container_type(target: object, selector: str) -> ContainerTypes:
    link_types = strip_annotation(get_annotations(target)[selector])

    if list in link_types:
        result = ContainerTypes.UNORDERED
    elif tuple in link_types:
        result = ContainerTypes.ORDERED
    elif any([PCBaseModel in link_type.mro() for link_type in link_types]):
        result = ContainerTypes.SINGLETON
    else:
        raise Exception(f"unexpected types {link_types}")

    return result


def get_item_link_container_type(item: ObjectIter.Entry) -> ContainerTypes:

    return get_link_container_type(item.target, item.selector)


class ChemComp(PCBaseModel):
    pass

    baseGlycoCode: Attribute[Optional[str]]  # 0..1
    beilsteinCode: Attribute[Optional[str]]  # 0..1
    caseRegCode: Attribute[Optional[str]]  # 0..1
    ccpCode: Attribute[str]  # 1..1
    code1Letter: Attribute[Optional[str]]  # 0..1
    code3Letter: Attribute[Optional[str]]  # 0..1
    commonNames: Attribute[OptionalOneOrMore[str]]  # 0..*
    details: Attribute[Optional[str]]  # 0..1
    hasStdChirality: Attribute[Optional[bool]]  # 0..1
    # isLinearPolymer: bool   derived                        # 1..1
    keywords: Attribute[OptionalOneOrMore[str]]  # 0..*
    merckCode: Attribute[Optional[str]]  # 0..1
    molType: Attribute[MolType]  # 1..1
    name: Attribute[Optional[str]]  # 0..1
    sigmaAldrichCode: Attribute[Optional[str]]  # 0..1
    stdChemCompCode: Attribute[Optional[str]]  # 0..1
    #
    # TODO: should be in parent class...
    guid: Attribute[str]
    createdBy: Attribute[str]
    #
    chemAngles: Children[OptionalOneOrMoreOfType[ChemAngle]]  # 0..*
    chemAtomSets: Children[OptionalOneOrMoreOfType[ChemAtomSet]]  # 0..*
    chemAtoms: Children[
        List[Annotated[Union[ChemAtom, LinkAtom], Field(discriminator="type")]]
    ]  # 0..*
    chemBonds: Children[OptionalOneOrMoreOfType[ChemBond]]  # 0..*
    chemCompVars: Children[OptionalOneOrMoreOfType[ChemCompVar]]  # 0..*
    chemTorsions: Children[OptionalOneOrMoreOfType[ChemTorsion]]  # 0..*
    #     memopsRoot: Parent[object]  # Parent[MemopsRoot] parent - not implemented # 1..1
    linkEnds: Children[OptionalOneOrMoreOfType[LinkEnd]]  # 0..*
    namingSystems: Children[OptionalOneOrMoreOfType[NamingSystem]]  # 0..*
    parent: Parent[object]  # : Parent[MemopsRoot]  # 1..1
    # residueTypeProbabilities - out of package not implemented                            # 0..*
    # stdChemComp  #derived                                                                # 0..1
    stereochemistries: Children[OptionalOneOrMoreOfType[Stereochemistry]]  # 0..*

    applicationData: Children[
        OptionalOneOrMore[Annotated[AppDataTypes, Field(discriminator="type")]]
    ]

    def __post_init__(self):
        self._make_links()

    def _make_links(self):

        object_iter = ObjectIter(result_generator=item_only)
        for_parent = []
        for_children = []

        for i, item in enumerate(object_iter.iter(self)):
            linkage = get_item_linkage(item)

            if linkage == Linkage.parent:
                for_parent.append(item)
            elif linkage == Linkage.link:
                if item.value is not None:
                    for_children.append(item)

        for item in for_parent:
            setattr(
                item.target,
                item.selector,
                object_iter.get_parent(object_iter.get_parent(item)).target,
            )

        for j, item in enumerate(for_children):

            value = item.value

            link_container_type = get_item_link_container_type(item)

            if link_container_type is ContainerTypes.SINGLETON:

                replacement_value = self.__class__._object_map[item.value]

                setattr(item.target, item.selector, replacement_value)

            elif link_container_type is ContainerTypes.ORDERED:

                new_value = [None] * len(value)
                for i, elem in enumerate(item.value):
                    if isinstance(elem, ID):
                        new_value[i] = self.__class__._object_map[elem]
                    elif isinstance(elem, PCBaseModel):
                        new_value[i] = elem
                    else:
                        raise Exception("unexpected")
                setattr(item.target, item.selector, tuple(new_value))
            elif link_container_type is ContainerTypes.UNORDERED:

                if isinstance(item.value, PCBaseModel):
                    setattr(item.target, item.selector, [item.value])
                elif isinstance(item.value, ID):
                    setattr(
                        item.target,
                        item.selector,
                        [self.__class__._object_map[item.value]],
                    )
                else:
                    for i, elem in enumerate(item.value):
                        if isinstance(elem, ID):
                            value[i] = self.__class__._object_map[elem]
                        elif isinstance(elem, PCBaseModel):
                            value[i] = elem
                        else:
                            raise Exception("unexpected")
            else:
                raise Exception("unexpected")

            # now iterate again setting opposite links...
            opposite_selector = get_item_opposite_selector(item)

            new_value = getattr(item.target, item.selector)

            if link_container_type is ContainerTypes.UNORDERED:
                if not isinstance(new_value, list):
                    raise Exception("unexpected")
            if link_container_type is ContainerTypes.ORDERED:
                if not isinstance(new_value, tuple):
                    raise Exception("unexpected")

            elif link_container_type in (
                ContainerTypes.UNORDERED,
                ContainerTypes.ORDERED,
            ):

                if len(new_value) > 0:

                    for elem in new_value:

                        self._ensure_opposite_container(elem, opposite_selector, item)

                        opposite_container_type = get_link_container_type(
                            new_value[0], opposite_selector
                        )

                        if opposite_container_type is ContainerTypes.SINGLETON:
                            self._setup_singleton_opposite(
                                elem, opposite_selector, item
                            )

                        elif opposite_container_type in (
                            ContainerTypes.UNORDERED,
                            ContainerTypes.ORDERED,
                        ):

                            self._setup_container_opposite(
                                elem, opposite_selector, opposite_container_type, item
                            )

    def _setup_container_opposite(
        self, elem, opposite_selector, opposite_container_type, item
    ):

        opposite_values = getattr(elem, opposite_selector)

        target_values = (
            list(opposite_values)
            if opposite_container_type is ContainerTypes.ORDERED
            else opposite_values
        )

        if len(target_values) > 0 and isinstance(target_values[0], ID):
            for i, target_value in enumerate(target_values):
                if isinstance(target_value, ID):
                    target_values[i] = self.__class__._object_map[target_value]

        # this looks a little odd but object comparison currently fails and needs looking at
        is_correct_id = [elem.ID == item.target.ID for elem in target_values]
        needs_append = not any(is_correct_id)

        if needs_append and opposite_container_type is ContainerTypes.ORDERED:
            raise Exception("unexpected for ordered container")
        if needs_append:
            raise Exception("had to append")
            # target_values.append(item.target)

        if opposite_container_type is ContainerTypes.ORDERED:
            setattr(elem, opposite_selector, tuple(target_values))

        # ids = [new_elem.ID for new_elem in getattr(elem, opposite_selector)]
        # if not item.target.ID in ids:
        #     raise Exception('Unexpected')

    def _setup_singleton_opposite(self, elem, opposite_selector, item):

        opposite_value = getattr(elem, opposite_selector)

        if isinstance(opposite_value, ID):
            new_opposite_value = self.__class__._object_map[opposite_value]
        elif opposite_value is None:
            new_opposite_value = item.target
        else:
            new_opposite_value = opposite_value

        setattr(elem, opposite_selector, new_opposite_value)

        if getattr(elem, opposite_selector).ID != item.target.ID:
            raise Exception(
                f"unexpected mismatch! {getattr(elem, opposite_selector).ID}, {item.target.ID}"
            )

    def _ensure_opposite_container(
        self, elem, opposite_selector, opposite_container_type
    ):
        opposite_value = getattr(elem, opposite_selector)

        need_list = (
            opposite_container_type is ContainerTypes.UNORDERED
            and not isinstance(opposite_value, list)
        )
        need_tuple = (
            opposite_container_type is ContainerTypes.ORDERED
            and not isinstance(opposite_value, tuple)
        )

        if need_list or need_tuple:

            new_opposite_value = []
            if opposite_value is not None:
                new_opposite_value.append(opposite_value)

                if need_tuple:
                    new_opposite_value = tuple(new_opposite_value)
                setattr(elem, opposite_selector, new_opposite_value)

    def json(self, original=False):

        if original:
            result = super(ChemComp, self).json()
        else:
            copy_of_data = self.model_copy(deep=True)
            copy_of_data.unlink()
            result = copy_of_data.json(original=True)
        return result

    def unlink(self):
        object_iter = ObjectIter(
            result_generator=item_only, child_lister=object_children_root_only
        )
        for_parent = []
        for_children = []
        for i, item in enumerate(object_iter.iter(self)):

            linkage = get_item_linkage(item)

            if linkage == Linkage.parent:
                for_parent.append(item)
            elif linkage == Linkage.link:
                if item.value is not None:
                    for_children.append(item)

        for item in for_parent:
            setattr(item.target, item.selector, None)
        for j, item in enumerate(for_children):

            value = item.value
            if isinstance(value, PCBaseModel):

                setattr(item.target, item.selector, value.ID)
            elif isinstance(value, tuple):

                new_value = [None] * len(value)
                for i, elem in enumerate(item.value):

                    if isinstance(elem, PCBaseModel):
                        new_value[i] = elem.ID
                setattr(item.target, item.selector, tuple(new_value))

            elif isinstance(value, list):

                for i, elem in enumerate(item.value):
                    if isinstance(elem, PCBaseModel):
                        value[i] = elem.ID

        # for object in self.__class__.__opposites__:
        #     print(object)



import typer

from nef_pipelines.lib.test_lib import (
    NOQA_E501,
    assert_lines_match,
    isolate_loop,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.tools.frames.unassign import unassign

EXIT_ERROR = 1

app = typer.Typer()
app.command()(unassign)

EXPECTED_BASIC_SHIFT_LIST = """
loop_
    _nef_chemical_shift.chain_code
    _nef_chemical_shift.sequence_code
    _nef_chemical_shift.residue_name
    _nef_chemical_shift.atom_name
    _nef_chemical_shift.value
    _nef_chemical_shift.value_uncertainty
    _nef_chemical_shift.element
    _nef_chemical_shift.isotope_number

    @-     @1     .    C     174.078751  .               C  13
    @-     @1     .    H       8.507345  0.0031493880    H   1
    @-     @1-1   .    C     175.947447  0.08720577414   C  13
    @-     @1-1   .    CA     56.689055  0.0499432144    C  13
    @-     @2     .    C      17.078751  .               C  13
    @-     @2     .    H      18.507345  0.0031493880    H   1
    @-     @2     .    N      23.365410  0.1052522259    N  15
    @-     @2-1   .    C      75.947447  0.08720577414   C  13
    @-     @2-1   .    CA      6.689055  0.0499432144    C  13
    @-     @2-1   .    CB      1.543048  0.1003766534    C  13
    @-     @3     .    C       7.090786  .               C  13
    @-     @3     .    N      12.714456  0.1199761464    N  15
    @-     @3-1   .    C      78.121295  0.3397487528    C  13
    @-     @3-1   .    CB      6.570846  0.0514681685    C  13
    @-     @4     .    H      88.111256  0.0019229051    H   1
    @-     @4     .    N     112.453575  0.0925289285    N  15
    @-     @4-1   .    C     167.261980  0.0133395288    C  13
    @-     @4-1   .    CB    116.427729  0.1141368283    C  13
stop_
"""

EXPECTED_BASIC_SPECTRUM = """
loop_                                          # noqa: E501
    _nef_peak.index
    _nef_peak.peak_id
    _nef_peak.volume
    _nef_peak.volume_uncertainty
    _nef_peak.height
    _nef_peak.height_uncertainty
    _nef_peak.position_1
    _nef_peak.position_uncertainty_1
    _nef_peak.position_2
    _nef_peak.position_uncertainty_2
    _nef_peak.position_3
    _nef_peak.position_uncertainty_3
    _nef_peak.chain_code_1
    _nef_peak.sequence_code_1
    _nef_peak.residue_name_1
    _nef_peak.atom_name_1
    _nef_peak.chain_code_2
    _nef_peak.sequence_code_2
    _nef_peak.residue_name_2
    _nef_peak.atom_name_2
    _nef_peak.chain_code_3
    _nef_peak.sequence_code_3
    _nef_peak.residue_name_3
    _nef_peak.atom_name_3

    1   0   0.38    .   0.38    .   10.405   .   8.796   .   132.49   .   @-   @5   .   HE1   @-   @4   .   HN   @-   @5   .   N
    2   1   1.298   .   1.298   .   10.408   .   7.139   .   132.49   .   .    .    .   .     .    .    .   .    .    .    .   .
    3   3   0.319   .   0.319   .   10.408   .   5.542   .   132.49   .   @-   @3   .   HE1   @-   @4   .   HA   .    .    .   .

stop_
""".replace(
    NOQA_E501, ""
)


# noinspection PyUnusedLocal
def test_unassign_basic():

    INPUT = open(
        path_in_test_data(__file__, "ubiquitin_short_unassign_single_chain.nef")
    ).read()
    result = run_and_report(app, [], input=INPUT)

    loop = isolate_loop(
        result.stdout, "nef_chemical_shift_list_default", "nef_chemical_shift"
    )
    assert_lines_match(EXPECTED_BASIC_SHIFT_LIST, loop)

    loop = isolate_loop(result.stdout, "nef_nmr_spectrum_simnoe", "nef_peak")
    assert_lines_match(EXPECTED_BASIC_SPECTRUM, loop)


EXPECTED_FULL_SHIFT_LIST = """
    loop_
        _nef_chemical_shift.chain_code
        _nef_chemical_shift.sequence_code
        _nef_chemical_shift.residue_name
        _nef_chemical_shift.atom_name
        _nef_chemical_shift.value
        _nef_chemical_shift.value_uncertainty
        _nef_chemical_shift.element
        _nef_chemical_shift.isotope_number

        @-     @1     .     .    174.078751  .               C  13
        @-     @1     .     .      8.507345  0.0031493880    H   1
        @-     @1-1   .     .    175.947447  0.08720577414   C  13
        @-     @1-1   .     .     56.689055  0.0499432144    C  13
        @-     @2     .     .     17.078751  .               C  13
        @-     @2     .     .     18.507345  0.0031493880    H   1
        @-     @2     .     .     23.365410  0.1052522259    N  15
        @-     @2-1   .     .     75.947447  0.08720577414   C  13
        @-     @2-1   .     .      6.689055  0.0499432144    C  13
        @-     @2-1   .     .      1.543048  0.1003766534    C  13
        @-     @3     .     .      7.090786  .               C  13
        @-     @3     .     .     12.714456  0.1199761464    N  15
        @-     @3-1   .     .     78.121295  0.3397487528    C  13
        @-     @3-1   .     .      6.570846  0.0514681685    C  13
        @-     @4     .     .     88.111256  0.0019229051    H   1
        @-     @4     .     .    112.453575  0.0925289285    N  15
        @-     @4-1   .     .    167.261980  0.0133395288    C  13
        @-     @4-1   .     .    116.427729  0.1141368283    C  13
    stop_
    """

EXPECTED_FULL_SPECTRUM = """
    loop_                                          # noqa: E501
        _nef_peak.index
        _nef_peak.peak_id
        _nef_peak.volume
        _nef_peak.volume_uncertainty
        _nef_peak.height
        _nef_peak.height_uncertainty
        _nef_peak.position_1
        _nef_peak.position_uncertainty_1
        _nef_peak.position_2
        _nef_peak.position_uncertainty_2
        _nef_peak.position_3
        _nef_peak.position_uncertainty_3
        _nef_peak.chain_code_1
        _nef_peak.sequence_code_1
        _nef_peak.residue_name_1
        _nef_peak.atom_name_1
        _nef_peak.chain_code_2
        _nef_peak.sequence_code_2
        _nef_peak.residue_name_2
        _nef_peak.atom_name_2
        _nef_peak.chain_code_3
        _nef_peak.sequence_code_3
        _nef_peak.residue_name_3
        _nef_peak.atom_name_3

        1   0   0.38    .   0.38    .   10.405   .   8.796   .   132.49   .   @-   @5   .   .     @-   @4   .   .     @-   @5   .   .
        2   1   1.298   .   1.298   .   10.408   .   7.139   .   132.49   .   .    .    .   .     .    .    .   .     .    .    .   .
        3   3   0.319   .   0.319   .   10.408   .   5.542   .   132.49   .   @-   @3   .   .     @-   @4   .   .     .    .    .   .

    stop_
""".replace(
    NOQA_E501, ""
)


def test_unassign_all():

    INPUT = open(
        path_in_test_data(__file__, "ubiquitin_short_unassign_single_chain.nef")
    ).read()
    result = run_and_report(app, ["--targets", "all"], input=INPUT)

    loop = isolate_loop(
        result.stdout, "nef_chemical_shift_list_default", "nef_chemical_shift"
    )
    assert_lines_match(loop, EXPECTED_FULL_SHIFT_LIST)

    loop = isolate_loop(result.stdout, "nef_nmr_spectrum_simnoe", "nef_peak")
    assert_lines_match(EXPECTED_FULL_SPECTRUM, loop)


EXPECTED_UNUSED_SHIFT_LIST = """
    loop_
        _nef_chemical_shift.chain_code
        _nef_chemical_shift.sequence_code
        _nef_chemical_shift.residue_name
        _nef_chemical_shift.atom_name
        _nef_chemical_shift.value
        _nef_chemical_shift.value_uncertainty
        _nef_chemical_shift.element
        _nef_chemical_shift.isotope_number

        .     .     .     .    174.078751  .               C  13
        .     .     .     .      8.507345  0.0031493880    H   1
        .     .     .     .    175.947447  0.08720577414   C  13
        .     .     .     .     56.689055  0.0499432144    C  13
        .     .     .     .     17.078751  .               C  13
        .     .     .     .     18.507345  0.0031493880    H   1
        .     .     .     .     23.365410  0.1052522259    N  15
        .     .     .     .     75.947447  0.08720577414   C  13
        .     .     .     .      6.689055  0.0499432144    C  13
        .     .     .     .      1.543048  0.1003766534    C  13
        .     .     .     .      7.090786  .               C  13
        .     .     .     .     12.714456  0.1199761464    N  15
        .     .     .     .     78.121295  0.3397487528    C  13
        .     .     .     .      6.570846  0.0514681685    C  13
        .     .     .     .     88.111256  0.0019229051    H   1
        .     .     .     .    112.453575  0.0925289285    N  15
        .     .     .     .    167.261980  0.0133395288    C  13
        .     .     .     .    116.427729  0.1141368283    C  13
    stop_
    """

EXPECTED_UNUSED_SPECTRUM = """
    loop_                                          # noqa: E501
        _nef_peak.index
        _nef_peak.peak_id
        _nef_peak.volume
        _nef_peak.volume_uncertainty
        _nef_peak.height
        _nef_peak.height_uncertainty
        _nef_peak.position_1
        _nef_peak.position_uncertainty_1
        _nef_peak.position_2
        _nef_peak.position_uncertainty_2
        _nef_peak.position_3
        _nef_peak.position_uncertainty_3
        _nef_peak.chain_code_1
        _nef_peak.sequence_code_1
        _nef_peak.residue_name_1
        _nef_peak.atom_name_1
        _nef_peak.chain_code_2
        _nef_peak.sequence_code_2
        _nef_peak.residue_name_2
        _nef_peak.atom_name_2
        _nef_peak.chain_code_3
        _nef_peak.sequence_code_3
        _nef_peak.residue_name_3
        _nef_peak.atom_name_3

        1   0   0.38    .   0.38    .   10.405   .   8.796   .   132.49   .   .   .   .   .   .   .   .   .   .   .   .   .
        2   1   1.298   .   1.298   .   10.408   .   7.139   .   132.49   .   .   .   .   .   .   .   .   .   .   .   .   .
        3   3   0.319   .   0.319   .   10.408   .   5.542   .   132.49   .   .   .   .   .   .   .   .   .   .   .   .   .

    stop_
""".replace(
    NOQA_E501, ""
)


def test_unassign_complete():

    INPUT = open(
        path_in_test_data(__file__, "ubiquitin_short_unassign_single_chain.nef")
    ).read()
    result = run_and_report(
        app, ["--targets", "all", "--sequence-mode", "unused"], input=INPUT
    )

    loop = isolate_loop(
        result.stdout, "nef_chemical_shift_list_default", "nef_chemical_shift"
    )
    assert_lines_match(loop, EXPECTED_UNUSED_SHIFT_LIST)

    loop = isolate_loop(result.stdout, "nef_nmr_spectrum_simnoe", "nef_peak")
    assert_lines_match(EXPECTED_UNUSED_SPECTRUM, loop)


EXPECTED_ORDERED_SHIFT_LIST = """
    loop_
        _nef_chemical_shift.chain_code
        _nef_chemical_shift.sequence_code
        _nef_chemical_shift.residue_name
        _nef_chemical_shift.atom_name
        _nef_chemical_shift.value
        _nef_chemical_shift.value_uncertainty
        _nef_chemical_shift.element
        _nef_chemical_shift.isotope_number

        @-     @1     .    C     174.078751  .               C  13
        @-     @1     .    H       8.507345  0.0031493880    H   1
        @-     @1-1   .    C     175.947447  0.08720577414   C  13
        @-     @1-1   .    CA     56.689055  0.0499432144    C  13
        @-     @2     .    C      17.078751  .               C  13
        @-     @2     .    H      18.507345  0.0031493880    H   1
        @-     @2     .    N      23.365410  0.1052522259    N  15
        @-     @2-1   .    C      75.947447  0.08720577414   C  13
        @-     @2-1   .    CA      6.689055  0.0499432144    C  13
        @-     @2-1   .    CB      1.543048  0.1003766534    C  13
        @-     @4     .    C       7.090786  .               C  13
        @-     @4     .    N      12.714456  0.1199761464    N  15
        @-     @4-1   .    C      78.121295  0.3397487528    C  13
        @-     @4-1   .    CB      6.570846  0.0514681685    C  13
        @-     @5     .    H      88.111256  0.0019229051    H   1
        @-     @5     .    N     112.453575  0.0925289285    N  15
        @-     @5-1   .    C     167.261980  0.0133395288    C  13
        @-     @5-1   .    CB    116.427729  0.1141368283    C  13
    stop_
    """

EXPECTED_ORDERED_SPECTRUM = """
    loop_                                               # noqa: E501
        _nef_peak.index
        _nef_peak.peak_id
        _nef_peak.volume
        _nef_peak.volume_uncertainty
        _nef_peak.height
        _nef_peak.height_uncertainty
        _nef_peak.position_1
        _nef_peak.position_uncertainty_1
        _nef_peak.position_2
        _nef_peak.position_uncertainty_2
        _nef_peak.position_3
        _nef_peak.position_uncertainty_3
        _nef_peak.chain_code_1
        _nef_peak.sequence_code_1
        _nef_peak.residue_name_1
        _nef_peak.atom_name_1
        _nef_peak.chain_code_2
        _nef_peak.sequence_code_2
        _nef_peak.residue_name_2
        _nef_peak.atom_name_2
        _nef_peak.chain_code_3
        _nef_peak.sequence_code_3
        _nef_peak.residue_name_3
        _nef_peak.atom_name_3

        1   0   0.38    .   0.38    .   10.405   .   8.796   .   132.49   .   @-   @3   .   HE1   @-   @5   .   HN   @-   @3   .   N
        2   1   1.298   .   1.298   .   10.408   .   7.139   .   132.49   .   .    .    .   .     .    .    .   .    .    .    .   .
        3   3   0.319   .   0.319   .   10.408   .   5.542   .   132.49   .   @-   @4   .   HE1   @-   @5   .   HA   .    .    .   .

    stop_
""".replace(
    NOQA_E501, ""
)


def test_unassign_ordered():

    INPUT = open(
        path_in_test_data(__file__, "ubiquitin_short_unassign_single_chain.nef")
    ).read()
    result = run_and_report(app, ["--sequence-mode", "ordered"], input=INPUT)

    loop = isolate_loop(
        result.stdout, "nef_chemical_shift_list_default", "nef_chemical_shift"
    )
    assert_lines_match(loop, EXPECTED_ORDERED_SHIFT_LIST)

    loop = isolate_loop(result.stdout, "nef_nmr_spectrum_simnoe", "nef_peak")
    assert_lines_match(EXPECTED_ORDERED_SPECTRUM, loop)


EXPECTED_PRESERVED_SHIFT_LIST = """
    loop_
        _nef_chemical_shift.chain_code
        _nef_chemical_shift.sequence_code
        _nef_chemical_shift.residue_name
        _nef_chemical_shift.atom_name
        _nef_chemical_shift.value
        _nef_chemical_shift.value_uncertainty
        _nef_chemical_shift.element
        _nef_chemical_shift.isotope_number

        @-     @PR_1     .    C     174.078751  .               C  13
        @-     @PR_1     .    H       8.507345  0.0031493880    H   1
        @-     @PR_1-1   .    C     175.947447  0.08720577414   C  13
        @-     @PR_1-1   .    CA     56.689055  0.0499432144    C  13
        @-     @PR_2     .    C      17.078751  .               C  13
        @-     @PR_2     .    H      18.507345  0.0031493880    H   1
        @-     @PR_2     .    N      23.365410  0.1052522259    N  15
        @-     @PR_2-1   .    C      75.947447  0.08720577414   C  13
        @-     @PR_2-1   .    CA      6.689055  0.0499432144    C  13
        @-     @PR_2-1   .    CB      1.543048  0.1003766534    C  13
        @-     @2     .    C       7.090786  .               C  13
        @-     @2     .    N      12.714456  0.1199761464    N  15
        @-     @2-1   .    C      78.121295  0.3397487528    C  13
        @-     @2-1   .    CB      6.570846  0.0514681685    C  13
        @-     @3     .    H      88.111256  0.0019229051    H   1
        @-     @3     .    N     112.453575  0.0925289285    N  15
        @-     @3-1   .    C     167.261980  0.0133395288    C  13
        @-     @3-1   .    CB    116.427729  0.1141368283    C  13
    stop_
    """

EXPECTED_PRESERVED_SPECTRUM = """
    loop_                                          # noqa: E501
        _nef_peak.index
        _nef_peak.peak_id
        _nef_peak.volume
        _nef_peak.volume_uncertainty
        _nef_peak.height
        _nef_peak.height_uncertainty
        _nef_peak.position_1
        _nef_peak.position_uncertainty_1
        _nef_peak.position_2
        _nef_peak.position_uncertainty_2
        _nef_peak.position_3
        _nef_peak.position_uncertainty_3
        _nef_peak.chain_code_1
        _nef_peak.sequence_code_1
        _nef_peak.residue_name_1
        _nef_peak.atom_name_1
        _nef_peak.chain_code_2
        _nef_peak.sequence_code_2
        _nef_peak.residue_name_2
        _nef_peak.atom_name_2
        _nef_peak.chain_code_3
        _nef_peak.sequence_code_3
        _nef_peak.residue_name_3
        _nef_peak.atom_name_3

        1   0   0.38    .   0.38    .   10.405   .   8.796   .   132.49   .   @-   @1   .   HE1   @-   @3   .   HN   @-   @1   .   N
        2   1   1.298   .   1.298   .   10.408   .   7.139   .   132.49   .   .    .    .   .     .    .    .   .    .    .    .   .
        3   3   0.319   .   0.319   .   10.408   .   5.542   .   132.49   .   @-   @2   .   HE1   @-   @3   .   HA   .    .    .   .

    stop_
""".replace(
    NOQA_E501, ""
)


def test_unassign_preserve():
    INPUT = open(
        path_in_test_data(__file__, "ubiquitin_short_unassign_single_chain.nef")
    ).read()
    result = run_and_report(app, ["--sequence-mode", "preserve"], input=INPUT)

    loop = isolate_loop(
        result.stdout, "nef_chemical_shift_list_default", "nef_chemical_shift"
    )
    assert_lines_match(loop, EXPECTED_PRESERVED_SHIFT_LIST)

    loop = isolate_loop(result.stdout, "nef_nmr_spectrum_simnoe", "nef_peak")
    assert_lines_match(EXPECTED_PRESERVED_SPECTRUM, loop)



import typer
from freezegun import freeze_time

from nef_pipelines.lib.nef_lib import NEF_METADATA, NEF_MOLECULAR_SYSTEM
from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.lib.util import get_version
from nef_pipelines.transcoders.nmrview.importers.sequence import sequence

app = typer.Typer()
app.command()(sequence)


EXPECTED_3AA = """\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   A   1   ALA   start    .   .
     2   A   2   ALA   middle   .   .
     3   A   3   ALA   end      .   .

   stop_

save_"""

EXPECTED_3AA10 = """\
save_nef_molecular_system
   _nef_molecular_system.sf_category   nef_molecular_system
   _nef_molecular_system.sf_framecode  nef_molecular_system

   loop_
      _nef_sequence.index
      _nef_sequence.chain_code
      _nef_sequence.sequence_code
      _nef_sequence.residue_name
      _nef_sequence.linking
      _nef_sequence.residue_variant
      _nef_sequence.cis_peptide

     1   A   10   ALA   start    .   .
     2   A   11   ALA   middle   .   .
     3   A   12   ALA   end      .   .

   stop_

save_"""


# noinspection PyUnusedLocal
def test_3aa():

    path = path_in_test_data(__file__, "3aa.seq")
    result = run_and_report(app, [path], input=HEADER)

    mol_sys_result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3AA, mol_sys_result)


# noinspection PyUnusedLocal
def test_3aa10():

    path = path_in_test_data(__file__, "3aa10.seq")
    result = run_and_report(app, [path], input=HEADER)

    mol_sys_result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3AA10, mol_sys_result)


HEADER = open(path_in_test_data(__file__, "test_header_entry.txt")).read()

EXPECTED_HEADER = f"""\
save_nef_nmr_meta_data
   _nef_nmr_meta_data.sf_category      nef_nmr_meta_data
   _nef_nmr_meta_data.sf_framecode     nef_nmr_meta_data
   _nef_nmr_meta_data.format_name      nmr_exchange_format
   _nef_nmr_meta_data.format_version   1.1
   _nef_nmr_meta_data.program_name     NEFPipelines
   _nef_nmr_meta_data.script_name      sequence.py
   _nef_nmr_meta_data.program_version  {get_version()}
   _nef_nmr_meta_data.creation_date    2012-01-14T12:00:01.123456
   _nef_nmr_meta_data.uuid             NEFPipelines-2012-01-14T12:00:01.123456-1043321819

   loop_
      _nef_run_history.run_number
      _nef_run_history.program_name
      _nef_run_history.program_version
      _nef_run_history.script_name

     1   NEFPipelines   0.0.1   header.py

   stop_

save_
"""


# noinspection PyUnusedLocal
@freeze_time("2012-01-14 12:00:01.123456")
def test_pipe_header(fixed_seed):

    path = path_in_test_data(__file__, "3aa10.seq")
    path_header = path_in_test_data(__file__, "test_header_entry.txt")
    result = run_and_report(app, ["--in", path_header, path])

    mol_sys_result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)
    meta_data_result = isolate_frame(result.stdout, "%s" % NEF_METADATA)

    assert_lines_match(EXPECTED_3AA10, mol_sys_result)
    assert_lines_match(EXPECTED_HEADER, meta_data_result)


# noinspection PyUnusedLocal
@freeze_time("2012-01-14 12:00:01.123456")
def test_header(fixed_seed):

    path = path_in_test_data(__file__, "3aa10.seq")
    result = run_and_report(app, [path], input=HEADER)

    mol_sys_result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3AA10, mol_sys_result)



import typer

from nef_pipelines.lib.nef_lib import NEF_MOLECULAR_SYSTEM
from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_frame,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.rcsb.importers.sequence import sequence

app = typer.Typer()
app.command()(sequence)

HEADER = open(path_in_test_data(__file__, "test_header_entry.txt")).read()

EXPECTED_3AA = """\
    save_nef_molecular_system
        _nef_molecular_system.sf_category   nef_molecular_system
        _nef_molecular_system.sf_framecode  nef_molecular_system

        loop_
            _nef_sequence.index
            _nef_sequence.chain_code
            _nef_sequence.sequence_code
            _nef_sequence.residue_name
            _nef_sequence.linking
            _nef_sequence.residue_variant
            _nef_sequence.cis_peptide

            1   A   1   ALA   start    .   .
            2   A   2   ALA   middle   .   .
            3   A   3   ALA   end      .   .

        stop_

    save_
"""


# noinspection PyUnusedLocal
def test_3aa():

    path = path_in_test_data(__file__, "3aa.pdb")
    result = run_and_report(app, [path], input=HEADER)

    assert result.exit_code == 0
    mol_sys_result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3AA, mol_sys_result)


EXPECTED_3A_AB = """\
    save_nef_molecular_system
        _nef_molecular_system.sf_category   nef_molecular_system
        _nef_molecular_system.sf_framecode  nef_molecular_system

        loop_
            _nef_sequence.index
            _nef_sequence.chain_code
            _nef_sequence.sequence_code
            _nef_sequence.residue_name
            _nef_sequence.linking
            _nef_sequence.residue_variant
            _nef_sequence.cis_peptide

            1   A    1   ALA   start    .   .
            2   A    2   ALA   middle   .   .
            3   A    3   ALA   end      .   .
            4   B   11   ALA   start    .   .
            5   B   12   ALA   middle   .   .
            6   B   13   ALA   end      .   .

        stop_

    save_
"""


# noinspection PyUnusedLocal
def test_3a_ab():

    path = path_in_test_data(__file__, "3a_ab.pdb")
    result = run_and_report(app, [path], input=HEADER)

    assert result.exit_code == 0

    mol_sys_result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3A_AB, mol_sys_result)


EXPECTED_3A_CCCC_DDDD = """\
    save_nef_molecular_system
        _nef_molecular_system.sf_category   nef_molecular_system
        _nef_molecular_system.sf_framecode  nef_molecular_system

        loop_
            _nef_sequence.index
            _nef_sequence.chain_code
            _nef_sequence.sequence_code
            _nef_sequence.residue_name
            _nef_sequence.linking
            _nef_sequence.residue_variant
            _nef_sequence.cis_peptide

            1   CCCC    1   ALA   start    .   .
            2   CCCC    2   ALA   middle   .   .
            3   CCCC    3   ALA   end      .   .
            4   DDDD   11   ALA   start    .   .
            5   DDDD   12   ALA   middle   .   .
            6   DDDD   13   ALA   end      .   .

        stop_

    save_"""


# noinspection PyUnusedLocal
def test_3a_segid_cccc_dddd():

    path = path_in_test_data(__file__, "3a_ab.pdb")
    result = run_and_report(app, ["--segid", path], input=HEADER)

    mol_sys_result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3A_CCCC_DDDD, mol_sys_result)


# noinspection PyUnusedLocal
def test_3a_force_segid_cccc_dddd():

    path = path_in_test_data(__file__, "3a_cccc_dddd.pdb")
    result = run_and_report(app, [path], input=HEADER)

    assert result.exit_code == 0

    mol_sys_result = isolate_frame(result.stdout, "%s" % NEF_MOLECULAR_SYSTEM)

    assert_lines_match(EXPECTED_3A_CCCC_DDDD, mol_sys_result)


# noinspection PyUnusedLocal
def test_3a_test_no_chain_segid():

    path = path_in_test_data(__file__, "3a_no_chain_no_segid.pdb")
    result = run_and_report(app, [path], input=HEADER, expected_exit_code=1)

    assert result.exit_code == 1

    assert "ERROR" in result.stdout
    assert "both the chain code and segment id" in result.stdout
    assert "not present on an ATOM record" in result.stdout



from typing import List

from nef_pipelines.lib.structures import SequenceResidue

PSF_FILE_IDENTIFIER = "PSF"  # text  expected on the first line of a psf file
PSF_NUM_NATOM_FIELDS = 2  # number of fields on the !NATOM line
PSF_NATOM_BLOCK_KEY_WORD = "!NATOM"  # idenifier for the !NATOM block
PSF_NATOM_COUNT_INDEX = 0  # index on the !NATOM line of the number of atoms
PSF_NATOM_KEY_WORD_INDEX = 1  # index on the !NATOM line of the !NATOM keyword
XPLOR_PSF_NUM_ATOM_FIELDS = 9  # number of fields in each NATOM block line

PSF_SEGID_INDEX = 1  # index of the segid in a line of the NATOM block
PSF_RESID_INDEX = 2  # index of the resid in a line of the NATOM block
PSF_ATOM_NAME_INDEX = 3  # index of the atom name  in a line of the NATOM block


def parse_xplor_PSF(text: str, file_name: str = "unknown") -> List[SequenceResidue]:
    in_data = False
    data_count = 0
    natom = None

    residues = set()
    for line_no, line in enumerate(text.split("\n"), start=1):
        if line_no == 1:
            if line.strip() != PSF_FILE_IDENTIFIER:
                msg = f"the first line of a PSF file should be '{PSF_FILE_IDENTIFIER}' i got '{line.rstrip()}'"
                raise_parse_exception(file_name, line_no, line, msg)

        if line_no > 1 and not in_data:
            fields = line.strip().split()
            if (
                len(fields) == PSF_NUM_NATOM_FIELDS
                and fields[PSF_NATOM_KEY_WORD_INDEX] == PSF_NATOM_BLOCK_KEY_WORD
            ):
                in_data = True
                try:
                    natom = int(fields[PSF_NATOM_COUNT_INDEX])
                except ValueError:
                    msg = "can't convert natom to an int"
                    raise_parse_exception(file_name, line_no, line, msg)

            continue

        if in_data and data_count >= natom:
            in_data = False
            break

        if in_data:
            data_count += 1
            fields = line.split()

            num_fields = len(fields)

            if num_fields != XPLOR_PSF_NUM_ATOM_FIELDS:
                msg = f"there are {num_fields} fields for atom number {data_count}; \
                        i expected {XPLOR_PSF_NUM_ATOM_FIELDS}"
                raise_parse_exception(file_name, line_no, line, msg)

            chain_code = fields[PSF_SEGID_INDEX]
            try:
                sequence_code = int(fields[PSF_RESID_INDEX])
            except ValueError:
                msg = f"i couldn't convert the residue number [field {PSF_RESID_INDEX}] to an integer, \
                      value was: {fields[PSF_RESID_INDEX]}"
                raise_parse_exception(file_name, line_no, line, msg)

            residue_type = fields[PSF_ATOM_NAME_INDEX]

            residue = SequenceResidue(
                chain_code=chain_code,
                sequence_code=sequence_code,
                residue_name=residue_type,
            )
            residues.add(residue)

    if natom != data_count:
        msg = (
            f" !unexpected! expected to read {natom} atoms but got {data_count} instead"
        )
        raise_parse_exception(file_name, line_no, line, msg)

    if not residues:
        msg = "no residues found!"
        raise_parse_exception(file_name, line_no, line, msg)

    residues = list(residues)
    residues.sort()

    return tuple(residues)


class PSFParseException(Exception):
    ...


def raise_parse_exception(file_name, line_no, line, specific_message):
    msg = f"""\
            Error: {specific_message}\
                at line {line_no}
                in file {file_name}
                line was "{line}"
        """
    raise PSFParseException(msg)



from argparse import Namespace
from inspect import currentframe, getargvalues, getouterframes


def get_args():
    frame = currentframe()
    outer_frames = getouterframes(frame)
    caller_frame = outer_frames[1][0]
    args = getargvalues(caller_frame).locals
    return Namespace(**args)



import typer

from nef_pipelines import nef_app

app = typer.Typer()
export_app = typer.Typer()
import_app = typer.Typer()

if nef_app.app:

    nef_app.app.add_typer(
        app, name="mars", help="- read and write mars [shifts and sequences]"
    )

    app.add_typer(
        export_app, name="export", help="- export mars [shifts and sequences]"
    )

    app.add_typer(import_app, name="import", help="- import mars [shifts]")

    # import of specific importers must be after app creation to avoid circular imports
    import nef_pipelines.transcoders.mars.exporters.fragments  # noqa: F401
    import nef_pipelines.transcoders.mars.exporters.input  # noqa: F401
    import nef_pipelines.transcoders.mars.exporters.sequence  # noqa: F401
    import nef_pipelines.transcoders.mars.exporters.shifts  # noqa: F401
    import nef_pipelines.transcoders.mars.importers.peaks  # noqa: F401
    import nef_pipelines.transcoders.mars.importers.shifts  # noqa: F401



from itertools import chain, cycle, islice, zip_longest
from pathlib import Path
from typing import List

import typer
from ordered_set import OrderedSet
from pynmrstar import Entry

from nef_pipelines.lib.nef_lib import read_or_create_entry_exit_error_on_bad_file
from nef_pipelines.lib.sequence_lib import (
    MoleculeType,
    chain_code_iter,
    offset_chain_residues,
    sequence_3let_to_res,
    sequence_to_nef_frame,
    translate_1_to_3,
)
from nef_pipelines.lib.structures import SequenceResidue
from nef_pipelines.lib.util import STDIN, exit_error, parse_comma_separated_options
from nef_pipelines.transcoders.sparky import import_app

app = typer.Typer()

NO_CHAIN_START_HELP = """don't include the start chain link type on a chain for the first residue [linkage will be
                         middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call this
                         option multiple times to set chain starts for multiple chains"""
NO_CHAIN_END_HELP = """don't include the end chain link type on a chain for the last residue [linkage will be
                       middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call this
                       option multiple times to set chain ends for multiple chains"""


# todo add comment to other sequences etc
@import_app.command()
def sequence(
    chain_codes: List[str] = typer.Option(
        [],
        "--chains",
        help="chain codes to use for the imported chains, can be a a comma sepatared list or can be called "
        "multiple times",
        metavar="<CHAIN-CODES>",
    ),
    starts: List[str] = typer.Option(
        [],
        "--starts",
        help="first residue number of sequences can be a comma separated list or ",
        metavar="<START>",
    ),
    no_chain_starts: List[str] = typer.Option(
        [], "--no-chain-start", help=NO_CHAIN_START_HELP
    ),
    no_chain_ends: List[str] = typer.Option(
        [], "--no-chain-end", help=NO_CHAIN_END_HELP
    ),
    entry_name: str = typer.Option("sparky", help="a name for the entry if required"),
    input: Path = typer.Option(
        STDIN,
        "-i",
        "--in",
        metavar="|PIPE|",
        help="pipe to read NEF data from, for testing [overrides stdin !use stdin instead!]",
    ),
    molecule_types: List[MoleculeType] = typer.Option(
        [
            MoleculeType.PROTEIN,
        ],
        "--molecule-type",
        help="molecule type",
    ),
    file_names: List[Path] = typer.Argument(
        ..., help="the file to read", metavar="<SPARKY-SEQUENCE-FILE>"
    ),
):
    """- convert sparky sequences to nef"""

    chain_codes = parse_comma_separated_options(chain_codes)
    if not chain_codes:
        chain_codes = ["A"]

    no_chain_starts = parse_comma_separated_options(no_chain_starts)
    if not no_chain_starts:
        no_chain_starts = [False]

    no_chain_ends = parse_comma_separated_options(no_chain_ends)
    if no_chain_ends:
        no_chain_ends = [False]

    starts = [int(elem) for elem in parse_comma_separated_options(starts)]
    if not starts:
        starts = [1]

    molecule_types = parse_comma_separated_options(molecule_types)
    if not molecule_types:
        molecule_types = [
            MoleculeType.PROTEIN,
        ]

    entry = read_or_create_entry_exit_error_on_bad_file(input, entry_name)

    entry = pipe(
        entry,
        file_names,
        chain_codes,
        starts,
        no_chain_starts,
        no_chain_ends,
        molecule_types,
    )

    print(entry)


def pipe(
    entry: Entry,
    file_names: List[Path],
    chain_codes: List[str],
    starts: List[int],
    no_chain_starts: List[bool],
    no_chain_ends: List[bool],
    molecule_types: List[MoleculeType],
):

    chain_code_iterator = chain_code_iter(chain_codes)

    chain_offsets = {}
    chains_with_no_start = set()
    chains_with_no_end = set()

    sparky_sequences = set()

    for (
        file_name,
        chain_code,
        molecule_type,
        start,
        no_chain_start,
        no_chain_end,
    ) in zip_longest(
        file_names,
        chain_code_iterator,
        molecule_types,
        starts,
        no_chain_starts,
        no_chain_ends,
    ):

        if file_name is None:
            break

        if molecule_type is None:
            molecule_type = MoleculeType.PROTEIN

        if start is None:
            start = 1

        if no_chain_start is None:
            no_chain_start = False

        if no_chain_end is None:
            no_chain_start = False

        if no_chain_start:
            chains_with_no_start.add(chain)

        if no_chain_end:
            chains_with_no_end.add(chain)

        chain_offsets[chain_code] = start

        sparky_sequences.update(read_sequence(file_name, chain_code, molecule_type))

    sparky_sequences = sorted(sparky_sequences)

    sparky_sequences = offset_chain_residues(sparky_sequences, chain_offsets)

    sparky_frame = sequence_to_nef_frame(
        sparky_sequences, no_chain_starts, no_chain_ends
    )

    entry.add_saveframe(sparky_frame)

    return entry


def _get_sequence_offsets(chain_codes: List[str], starts: List[int]):

    offsets = [start - 1 for start in starts]
    cycle_starts = chain(
        offsets,
        cycle(
            [
                0,
            ]
        ),
    )
    offsets = list(islice(cycle_starts, len(chain_codes)))

    return {chain_code: offset for chain_code, offset in zip(chain_codes, offsets)}


def residues_to_chain_codes(residues: List[SequenceResidue]) -> List[str]:
    return list(OrderedSet([residue.chain_code for residue in residues]))


# could do with taking a list of offsets
# noinspection PyUnusedLocal
def read_sequence(
    file_name: Path, chain_code: str, molecule_type: MoleculeType
) -> List[SequenceResidue]:

    try:
        with open(file_name) as handle:
            try:
                sequence = "".join(handle.readlines())
                sequence = "".join(sequence.split())
            except Exception as e:
                exit_error(f"Error reading sparky sequence file {str(file_name)}", e)

            sequence_3_let = translate_1_to_3(sequence, molecule_type=molecule_type)
            chain_residues = sequence_3let_to_res(sequence_3_let, chain_code)

    except IOError as e:
        exit_error(f"couldn't open {file_name} because:\n{e}", e)

    return chain_residues



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.sparky.exporters.peaks import peaks

SHIFTS_SPARKY = "nef_chemical_shift_list_sparky"

app = typer.Typer()
app.command()(peaks)

EXPECTED = """\
****************************** nef_nmr_spectrum_peaks ******************************

Assignment       w1       w2      Height  Volume

  PR_36N-H       120.519  6.169        9       1
  PR_66N-H       104.406  6.180       10       2
  PR_65N-H       104.408  6.517       11       3
  PR_67N-H       103.504  7.048       12       4

************************************************************************************

"""


def test_ppm_out_short():

    STREAM = open(path_in_test_data(__file__, "ubi_peaks_short.neff")).read()

    result = run_and_report(app, ["--file-name-template", "-"], input=STREAM)

    assert_lines_match(EXPECTED, result.stdout)


EXPECTED_SUPPRESS_ASSIGNMENT = """\
****************************** nef_nmr_spectrum_peaks ******************************

       w1       w2      Height  Volume

  120.519  6.169        9       1
  104.406  6.180       10       2
  104.408  6.517       11       3
  103.504  7.048       12       4
************************************************************************************

"""


def test_ppm_out_short_suppress_assigment():

    STREAM = open(path_in_test_data(__file__, "ubi_peaks_short.neff")).read()

    result = run_and_report(
        app,
        ["--file-name-template", "-", "--suppress-column", "assignment"],
        input=STREAM,
    )

    assert_lines_match(EXPECTED_SUPPRESS_ASSIGNMENT, result.stdout)


EXPECTED_NO_HEIGHT = """\
****************************** nef_nmr_spectrum_peaks ******************************

Assignment       w1       w2        Volume

  PR_36N-H       120.519  6.169          1
  PR_66N-H       104.406  6.180          2
  PR_65N-H       104.408  6.517          3
  PR_67N-H       103.504  7.048          4

************************************************************************************

"""


def test_ppm_out_short_no_height():

    STREAM = open(path_in_test_data(__file__, "ubi_peaks_short.neff")).read()

    result = run_and_report(
        app, ["--file-name-template", "-", "--suppress-column", "height"], input=STREAM
    )

    assert_lines_match(EXPECTED_NO_HEIGHT, result.stdout)


EXPECTED_NO_VOLUME = """\
****************************** nef_nmr_spectrum_peaks ******************************

Assignment       w1       w2      Height

  PR_36N-H       120.519  6.169        9
  PR_66N-H       104.406  6.180       10
  PR_65N-H       104.408  6.517       11
  PR_67N-H       103.504  7.048       12

************************************************************************************

"""


def test_ppm_out_short_no_volume():

    STREAM = open(path_in_test_data(__file__, "ubi_peaks_short.neff")).read()

    result = run_and_report(
        app, ["--file-name-template", "-", "--suppress-column", "volume"], input=STREAM
    )

    assert_lines_match(EXPECTED_NO_VOLUME, result.stdout)


EXPECTED_NO_HEIGHT_OR_VOLUME = """\
****************************** nef_nmr_spectrum_peaks ******************************

Assignment       w1       w2

  PR_36N-H       120.519  6.169
  PR_66N-H       104.406  6.180
  PR_65N-H       104.408  6.517
  PR_67N-H       103.504  7.048

************************************************************************************

"""


def test_ppm_out_short_no_height_or_volume():

    STREAM = open(path_in_test_data(__file__, "ubi_peaks_short.neff")).read()

    result = run_and_report(
        app,
        [
            "--file-name-template",
            "-",
            "--suppress-column",
            "volume",
            "--suppress-column",
            "height",
        ],
        input=STREAM,
    )

    assert_lines_match(EXPECTED_NO_HEIGHT_OR_VOLUME, result.stdout)


def test_ppm_out_short_no_height_or_volume_with_data():

    STREAM = open(path_in_test_data(__file__, "ubi_peaks_short.neff")).read()

    result = run_and_report(
        app,
        [
            "--file-name-template",
            "-",
            "--suppress-column",
            "volume,height",
            "--add-data",
        ],
        input=STREAM,
    )

    assert_lines_match(EXPECTED_NO_HEIGHT_OR_VOLUME, result.stdout)


EXPECTED_DIFFERENT_RESIDUES = """\
****************************** nef_nmr_spectrum_peaks ******************************

Assignment         w1       w2      Height  Volume

PR_67H-PR_68HA   7.856   4.049           3       1
PR_66H-HA        7.406   4.180          10       5

************************************************************************************

"""


def test_ppm_out_short_different_residues():

    STREAM = open(
        path_in_test_data(__file__, "ubi_peaks_short_different_residues.neff")
    ).read()

    result = run_and_report(app, ["--file-name-template", "-"], input=STREAM)

    assert_lines_match(EXPECTED_DIFFERENT_RESIDUES, result.stdout)


EXPECTED_CHAINS = """\
****************************** nef_nmr_spectrum_peaks ******************************

Assignment         w1       w2      Height  Volume

A.G66H-HA        7.806   4.280          12       7
A.G66H-A.G67HA   7.906   4.380          14       9

************************************************************************************
"""


def test_ppm_out_short_chains():

    STREAM = open(path_in_test_data(__file__, "ubi_peaks_short_chains.neff")).read()

    result = run_and_report(app, ["--file-name-template", "-"], input=STREAM)

    assert_lines_match(EXPECTED_CHAINS, result.stdout)


EXPECTED_NO_CHAINS = """\
****************************** nef_nmr_spectrum_peaks ******************************

Assignment         w1       w2      Height  Volume

G66H-HA         7.806    4.280          12       7
G66H-G67HA      7.906    4.380          14       9

************************************************************************************
"""


def test_ppm_out_short_no_chains():

    STREAM = open(path_in_test_data(__file__, "ubi_peaks_short_chains.neff")).read()

    result = run_and_report(
        app, ["--file-name-template", "-", "--no-chains"], input=STREAM
    )

    assert_lines_match(EXPECTED_NO_CHAINS, result.stdout)


EXPECTED_WITH_DATA = """\
****************************** nef_nmr_spectrum_peaks ******************************

Assignment       w1       w2      Data  Height  Volume

  PR_36N-H       120.519  6.169              9       1
  PR_66N-H       104.406  6.180             10       2
  PR_65N-H       104.408  6.517             11       3
  PR_67N-H       103.504  7.048             12       4

************************************************************************************

"""


def test_ppm_out_short_with_data():

    STREAM = open(path_in_test_data(__file__, "ubi_peaks_short.neff")).read()

    result = run_and_report(
        app, ["--file-name-template", "-", "--add-data"], input=STREAM
    )

    assert_lines_match(EXPECTED_WITH_DATA, result.stdout)


EXPECTED_FULL_ASSIGNMENTS = """\
****************************** nef_nmr_spectrum_peaks ******************************

Assignment       w1       w2      Height  Volume

  PR_36N-PR_36H  120.519  6.169        9       1
  PR_66N-PR_66H  104.406  6.180       10       2
  PR_65N-PR_65H  104.408  6.517       11       3
  PR_67N-PR_67H  103.504  7.048       12       4

************************************************************************************

"""


def test_ppm_out_short_full_assignments():

    STREAM = open(path_in_test_data(__file__, "ubi_peaks_short.neff")).read()

    result = run_and_report(
        app, ["--file-name-template", "-", "--full-assignments"], input=STREAM
    )

    assert_lines_match(EXPECTED_FULL_ASSIGNMENTS, result.stdout)


EXPECTED_NO_ASSIGNMENTS = """\
****************************** nef_nmr_spectrum_peaks ******************************

Assignment       w1     w2   Height  Volume

       ?-?  120.519  6.169        9       1
       ?-?  104.406  6.180       10       2
       ?-?  104.408  6.517       11       3
       ?-?  103.504  7.048       12       4

************************************************************************************

"""


def test_ppm_out_short_discard_assignments():

    STREAM = open(path_in_test_data(__file__, "ubi_peaks_short.neff")).read()

    result = run_and_report(
        app, ["--file-name-template", "-", "--discard-assignments"], input=STREAM
    )

    assert_lines_match(EXPECTED_NO_ASSIGNMENTS, result.stdout)


EXPECTED_WITH_FULL_ASSIGNMENTS = """\
****************************** nef_nmr_spectrum_peaks ******************************

Assignment       w1       w2      Height  Volume

  A.A36N-H    120.519  6.169           9       1
  A.G66N-H    104.406  6.180          10       2
  A.L65N-H    104.408  6.517          11       3
  A.V67N-H    103.504  7.048          12       4

************************************************************************************


"""


def test_ppm_out_short_with_full_assignments():

    STREAM = open(
        path_in_test_data(__file__, "peaks_short_with_assignments.neff")
    ).read()

    result = run_and_report(app, ["--file-name-template", "-"], input=STREAM)

    assert_lines_match(EXPECTED_WITH_FULL_ASSIGNMENTS, result.stdout)


EXPECTED_WITH_DIFFERENT_CHAIN_SEPARATOR = """\

****************************** nef_nmr_spectrum_peaks ******************************

Assignment       w1       w2      Height  Volume

  A_A36N-H    120.519  6.169           9       1
  A_G66N-H    104.406  6.180          10       2
  A_L65N-H    104.408  6.517          11       3
  A_V67N-H    103.504  7.048          12       4

************************************************************************************

"""


def test_ppm_out_short_with_different_chain_separator():

    STREAM = open(
        path_in_test_data(__file__, "peaks_short_with_assignments.neff")
    ).read()

    result = run_and_report(
        app, ["--file-name-template", "-", "--chain-separator", "_"], input=STREAM
    )

    assert_lines_match(EXPECTED_WITH_DIFFERENT_CHAIN_SEPARATOR, result.stdout)


EXPECTED_3D_m1 = """\
****************************** nef_nmr_spectrum_peaks ******************************

Assignment                     w1       w2       w3      Height  Volume

  PR_36N-H-CA               120.519  6.169    56.120          9       1
  PR_36N-PR_36H-PR_36CAm1   104.406  6.180    52.670         10       2
  PR_65N-H-CA               104.408  6.517    53.150         11       3
  PR_65N-PR_65H-PR_65CAm1   103.504  7.048    52.130         12       4

************************************************************************************


"""


def test_ppm_out_short_3d_m1():

    STREAM = open(
        path_in_test_data(__file__, "ubi_peaks_short_i_minus_one.neff")
    ).read()

    result = run_and_report(
        app,
        [
            "--file-name-template",
            "-",
        ],
        input=STREAM,
    )

    assert_lines_match(EXPECTED_3D_m1, result.stdout)


EXPECTED_3D_m1_no_negatives = """\
****************************** nef_nmr_spectrum_peaks ******************************

Assignment                      w1      w2       w3     Height  Volume

  PR_36N-H-CA               120.519  6.169   56.120          9       1
  PR_36N-PR_36H-?CA         104.406  6.180   52.670         10       2
  PR_65N-H-CA               104.408  6.517   53.150         11       3
  PR_65N-PR_65H-?CA         103.504  7.048   52.130         12       4

************************************************************************************


"""


def test_ppm_out_short_3d_m1_no_negatives():

    STREAM = open(
        path_in_test_data(__file__, "ubi_peaks_short_i_minus_one.neff")
    ).read()

    result = run_and_report(
        app, ["--file-name-template", "-", "--no-negative-residues"], input=STREAM
    )

    assert_lines_match(EXPECTED_3D_m1_no_negatives, result.stdout)


def test_ppm_out_short_overlapped_chains():

    STREAM = open(
        path_in_test_data(__file__, "ubi_peaks_short_overlapping_chains.neff")
    ).read()

    result = run_and_report(app, ["--no-chains"], input=STREAM, expected_exit_code=1)

    assert "ignore chains" in result.stdout
    assert " A " in result.stdout
    assert " B " in result.stdout
    assert "66" in result.stdout
    assert "36" in result.stdout



from pathlib import Path
from typing import List

import typer
from pynmrstar import Entry, Loop

from nef_pipelines.lib.nef_lib import read_entry_from_file_or_stdin_or_exit_error
from nef_pipelines.lib.sequence_lib import sequence_from_entry_or_exit
from nef_pipelines.lib.structures import Residue, SequenceResidue
from nef_pipelines.lib.util import STDIN, parse_comma_separated_options
from nef_pipelines.tools.frames import frames_app

FRAMES_HELP = """\
    the names of the  frames to use, this can be a comma separated list of name or the option can be called
    called multiple times. Wild cards are allowed. If no match is found wild cards are checked as well unless
    --exact is set
"""


@frames_app.command()
def filter(
    frame_selectors: List[str] = typer.Option(["default"], help=FRAMES_HELP),
    input: Path = typer.Option(
        STDIN,
        "--in",
        "-i",
        metavar="|INPUT|",
        help="input to read NEF data from [stdin = -]",
    ),
    unassigned: bool = typer.Option(
        False,
        help="if set lines which do not contain complete chemical shift assignments are deleted",
    ),
):
    """-  filter the lines in one or more save frames"""

    frame_selectors = parse_comma_separated_options(frame_selectors)

    # TODO add a check that we have some shift frames

    entry = read_entry_from_file_or_stdin_or_exit_error(input)

    entry = pipe(entry, frame_selectors, unassigned)

    print(entry)


# def _update_spectra_with_groups(spectra):
#     if ExperimentType.TRIPLE in spectra:
#         spectra.update(TRIPLE_SPECTRA)
#         spectra.remove(ExperimentType.TRIPLE)
#
#     return spectra
#
#
# def _exit_bad_spectrum_type(spectrum, known_spectra):
#     msg = """
#         the spectrum type {spectrum} is not know, available spectrum types are
#
#         {spectrum_types}
#     """
#     msg = dedent(msg)
#
#     spectrum_types = strings_to_tabulated_terminal_sensitive(known_spectra)
#     spectrum_types = indent(spectrum_types, FOUR_SPACES)
#
#     exit_error(f(msg))
#
#


def pipe(
    entry: Entry,
    frame_selectors: List[str],
    unassigned=False,
) -> Entry:

    if unassigned:
        sequence = sequence_from_entry_or_exit(entry)

        sequence_set = set(
            [Residue.from_sequence_residue(residue) for residue in sequence]
        )

        all_frames = entry.frame_dict.values()

        for frame in all_frames:
            for loop in frame.loops:

                rows_to_remove = set()
                tags = loop.tags

                loop_has_tags = {}
                for tag_name in (
                    "chain_code",
                    "sequence_code",
                    "residue_name",
                    "atom_name",
                ):
                    loop_has_tags[tag_name] = [tag.startswith(tag_name) for tag in tags]

                are_assignables = [any(has_tag) for has_tag in loop_has_tags.values()]
                is_assignable = all(are_assignables)

                tag_sets_by_end = {}
                if is_assignable:
                    for tag_name in (
                        "chain_code",
                        "sequence_code",
                        "residue_name",
                        "atom_name",
                    ):
                        for tag, is_chain_code in zip(tags, loop_has_tags[tag_name]):
                            if is_chain_code:
                                end = tag[len(tag_name) :]
                                tag_and_index = (tag, loop.tag_index(tag))
                                tag_sets_by_end.setdefault(end, []).append(
                                    tag_and_index
                                )

                    for i, row in enumerate(loop):
                        chain_code = None
                        sequence_code = None
                        residue_name = None
                        atom_name = None
                        for tag_set in tag_sets_by_end.values():
                            for tag, tag_index in tag_set:
                                value = row[tag_index]

                                if tag.startswith("chain_code"):
                                    chain_code = value
                                elif tag.startswith("sequence_code"):
                                    sequence_code = value
                                elif tag.startswith("residue_name"):
                                    residue_name = value
                                elif tag.startswith("atom_name"):
                                    atom_name = value

                        residue = Residue(
                            chain_code=chain_code,
                            sequence_code=sequence_code,
                            residue_name=residue_name,
                        )

                        residue_in_sequence = residue in sequence_set

                        # TODO:  this is a hack it needs proper atom identification
                        atom_name_ok = atom_name is not UNUSED

                        if not atom_name_ok or not residue_in_sequence:
                            rows_to_remove.add(i)

                for row_index in reversed(sorted(rows_to_remove)):
                    del loop.data[row_index]
    return entry


#
#     selected_shift_list_frames = select_frames_by_name(
#         all_shift_list_frames, shift_frame_selectors, exact=exact_frame_selectors
#     )
#
#     shifts = nef_frames_to_shifts(selected_shift_list_frames)
#
#     for spectrum in spectra:
#
#         spectrum = spectrum.lower()
#
#         if spectrum not in EXPERIMENT_INFO:
#             _exit_bad_spectrum_type(spectrum, list(EXPERIMENT_INFO.keys()))
#         spectrum_info = EXPERIMENT_INFO[spectrum]
#
#         peaks = make_spectrum(shifts, spectrum_info)
#
#         if len(peaks) == 0:
#             continue
#
#         frame_name = f(name_template_string)
#
#         dimensions = [
#             {"axis_code": dimension} for dimension in spectrum_info.dimensions
#         ]
#
#         # TODO add info about frames for errors
#         spectrum_classification = SPECTRUM_TYPE_TO_CLASSIFICATION[spectrum]
#         extra_tags = {
#             EXPERIMENT_CLASSIFICATION: spectrum_classification,
#             EXPERIMENT_TYPE: EXPERIMENT_CLASSIFICATION_TO_SYNONYM[
#                 spectrum_classification
#             ],
#         }
#         frame = peaks_to_frame(
#             peaks, dimensions, spectrometer_frequency, frame_name, extra_tags=extra_tags
#         )
#
#         entry.add_saveframe(frame)
#
#     return entry
#
#
# def make_spectrum(
#     shifts: List[ShiftData], info: PeakInfo, height: int = 1_000_000
# ) -> List[NewPeak]:
#
#     # note this is not very sophisticated and doesn't cover side-chains... we ought to port the ccpn code
#     relevant_shifts = []
#
#     for atom_name in info.atoms:
#         for shift in shifts:
#             shift_atom_name = shift.atom.atom_name.upper()
#             shift_atom_name = shift_atom_name.strip("1")
#             shift_atom_name = shift_atom_name.strip("-")
#
#             if shift_atom_name in atom_name:
#                 relevant_shifts.append(shift)
#
#     shifts_by_residue = {}
#
#     for shift in relevant_shifts:
#         residue = str(shift.atom.residue.sequence_code).split("-")[0]
#         shifts_by_residue.setdefault(residue, []).append(shift)
#
#     peaks = []
#
#     for residue_shifts in shifts_by_residue.values():
#
#         for atom_set, sign in zip(info.atom_sets, info.atom_set_signs):
#
#             atom_set_shifts = []
#             for number_sequence_code_fields_required, atom_name_required in atom_set:
#                 for shift in residue_shifts:
#
#                     # these are errors for the spectra we are interested in currently
#                     # maybe add a warning?
#                     if "+" in str(shift.atom.residue.sequence_code):
#                         continue
#
#                     number_sequence_code_fields = len(
#                         str(shift.atom.residue.sequence_code).split("-")
#                     )
#                     atom_name = shift.atom.atom_name
#
#                     number_fields_ok = (
#                         number_sequence_code_fields
#                         == number_sequence_code_fields_required
#                     )
#                     atom_name_ok = atom_name_required in atom_name
#
#                     if number_fields_ok and atom_name_ok:
#                         atom_set_shifts.append(shift)
#
#             if len(atom_set_shifts) == len(info.atoms):
#                 new_peak = NewPeak(
#                     atom_set_shifts, height=height * sign, volume=height * sign
#                 )
#                 peaks.append(new_peak)
#
#     return peaks



from argparse import Namespace
from pathlib import Path
from typing import List

import typer

from nef_pipelines.lib.sequence_lib import sequence_to_nef_frame
from nef_pipelines.lib.structures import SequenceResidue
from nef_pipelines.lib.typer_utils import get_args
from nef_pipelines.lib.util import (
    exit_error,
    parse_comma_separated_options,
    process_stream_and_add_frames,
)
from nef_pipelines.transcoders.rcsb.rcsb_lib import parse_pdb, parse_cif, guess_cif_or_pdb, RCSBFileType

from nef_pipelines.transcoders.rcsb import import_app

app = typer.Typer()

CHAINS_HELP = """chains [or segid] to read', metavar='<CHAIN-CODE>, multiple chains can be added as a comma joined list
                [eg A.B.C] ] or by calling this option mutiple times"""
NO_CHAIN_START_HELP = """don't include the start chain link type on a chain for the first residue [linkage will be
                         middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call this
                         option multiple times to set chain starts for multiple chains"""
NO_CHAIN_END_HELP = """don't include the end chain link type on a chain for the last residue [linkage will be
                       middle] for the named chains. Either use a comma joined list of chains [e.g. A,B] or call this
                       option multiple times to set chain ends for multiple chains"""


# noinspection PyUnusedLocal
@import_app.command(no_args_is_help=True)
def sequence(
    chain_codes: List[str] = typer.Option(None, "-c", "--chains", help=CHAINS_HELP),
    use_segids: bool = typer.Option(
        False, "-s", "--segid", help="use segid instead of chain"
    ),
    no_chain_starts: List[str] = typer.Option(
        [], "--no-chain-start", help=NO_CHAIN_START_HELP
    ),
    no_chain_ends: List[str] = typer.Option(
        [], "--no-chain-end", help=NO_CHAIN_END_HELP
    ),
    entry_name: str = typer.Option(
        "pdb", help="a name for the nef entry if not updating an existing entry"
    ),
    file_name: Path = typer.Argument(..., help="input pdb file", metavar="<PDB-FILE>"),
):
    """extracts a sequence from a pdb file"""

    chain_codes = parse_comma_separated_options(chain_codes)
    no_chain_starts = parse_comma_separated_options(no_chain_starts)
    no_chain_ends = parse_comma_separated_options(no_chain_ends)

    try:
        args = get_args()

        process_sequence(args)

    except Exception as e:
        exit_error(f"reading sequence failed because {e}", e)


def process_sequence(args: Namespace):

    pdb_sequences = read_sequences(
        args.file_name, args.chain_codes, use_segids=args.use_segids
    )

    if len(pdb_sequences) == 0:
        exit_error(f"no chains read from {args.file_name}")

    pdb_frame = sequence_to_nef_frame(
        pdb_sequences, set(args.no_chain_starts), set(args.no_chain_ends)
    )

    # TODO: need a a warning if the sequence already exists in a molecular system and ability to merge
    entry = process_stream_and_add_frames(
        [
            pdb_frame,
        ],
        args,
    )

    print(entry)


def read_sequences(path, target_chain_codes, use_segids=False):

    file_lines =  list(open(path).readlines())
    file_type = guess_cif_or_pdb(file_lines, str(path))

    if file_type is RCSBFileType.PDB:
        model = parse_pdb(file_lines)[0]
    elif file_type is RCSBFileType.CIF:
        model = parse_cif(file_lines)[0]
    else:
        msg  =  \
        f'''
            Couldn't determine if the file {path} was a cif or pdb file...
            are you sure the file has the right format?
        '''
        exit_error(msg)

    sequences = []


    if not use_segids:

        for chain in model:
            id = chain.chain_code
            if not id or len(id) == 0:
                use_segids = True

    all_chains = len(target_chain_codes) == 0


    for chain in model:
        for residue in chain:
            chain_code = chain.segment_id if use_segids else chain.chain_code
            chain_code = chain_code.strip()

            if not all_chains and chain_code not in target_chain_codes:
                continue

            sequence_code =  residue.sequence_code
            #TODO support a hetero atom flag
            # if len(hetero_atom_flag.strip()) != 0:
            #     continue

            if chain_code == "":
                exit_error(
                    f"residue with no chain code found for file {path} sequence_code is {sequence_code} \
                    residue_name is {residue.get_resname()}"
                )
            residue = SequenceResidue(
                chain_code=chain_code,
                sequence_code=sequence_code,
                residue_name=residue.residue_name,
            )
            sequences.append(residue)

    return sequences



import typer

from nef_pipelines.lib.test_lib import (
    assert_lines_match,
    isolate_loop,
    path_in_test_data,
    run_and_report,
)
from nef_pipelines.transcoders.mars.importers.peaks import peaks

app = typer.Typer()
app.command()(peaks)

NOQUA_E501 = "# noqa E501"
EXPECTED_HNCA = """
# noqa E501
loop_
    _nef_peak.index
    _nef_peak.peak_id
    _nef_peak.chain_code_1
    _nef_peak.sequence_code_1
    _nef_peak.residue_name_1
    _nef_peak.atom_name_1
    _nef_peak.chain_code_2
    _nef_peak.sequence_code_2
    _nef_peak.residue_name_2
    _nef_peak.atom_name_2
    _nef_peak.chain_code_3
    _nef_peak.sequence_code_3
    _nef_peak.residue_name_3
    _nef_peak.atom_name_3
    _nef_peak.position_1
    _nef_peak.position_uncertainty_1
    _nef_peak.position_2
    _nef_peak.position_uncertainty_2
    _nef_peak.position_3
    _nef_peak.position_uncertainty_3
    _nef_peak.height
    _nef_peak.height_uncertainty
    _nef_peak.volume
    _nef_peak.volume_uncertainty
    _nef_peak.ccpn_comment
    _nef_peak.ccpn_figure_of_merit


    0   0   A   1   MET   CA   A   2   GLN   N   A   2   GLN   H   54.608   .   122.899   .   8.907   .   .   .   .   .      merit=M   0.75
    1   1   A   2   GLN   CA   A   2   GLN   N   A   2   GLN   H   55.215   .   122.899   .   8.907   .   .   .   .   .      merit=M   0.75
    2   2   A   2   GLN   CA   A   3   ILE   N   A   3   ILE   H   55.057   .   115.127   .   8.318   .   .   .   .   .      merit=M   0.75
    3   3   A   3   ILE   CA   A   3   ILE   N   A   3   ILE   H   59.616   .   115.127   .   8.318   .   .   .   .   .      merit=M   0.75
    4   4   A   3   ILE   CA   A   4   PHE   N   A   4   PHE   H   59.599   .   118.73    .   8.614   .   .   .   .   .      merit=H   1.0
    5   5   A   4   PHE   CA   A   4   PHE   N   A   4   PHE   H   55.138   .   118.73    .   8.614   .   .   .   .   .      merit=H   1.0
    6   6   A   4   PHE   CA   A   5   VAL   N   A   5   VAL   H   55.142   .   121.424   .   9.307   .   .   .   .   .      merit=H   1.0
    7   7   A   5   VAL   CA   A   5   VAL   N   A   5   VAL   H   60.412   .   121.424   .   9.307   .   .   .   .   .      merit=H   1.0

stop_
""".replace(
    NOQUA_E501, ""
)

EXPECTED_HNcoCA = """
# noqa E501
loop_
    _nef_peak.index
    _nef_peak.peak_id
    _nef_peak.chain_code_1
    _nef_peak.sequence_code_1
    _nef_peak.residue_name_1
    _nef_peak.atom_name_1
    _nef_peak.chain_code_2
    _nef_peak.sequence_code_2
    _nef_peak.residue_name_2
    _nef_peak.atom_name_2
    _nef_peak.chain_code_3
    _nef_peak.sequence_code_3
    _nef_peak.residue_name_3
    _nef_peak.atom_name_3
    _nef_peak.position_1
    _nef_peak.position_uncertainty_1
    _nef_peak.position_2
    _nef_peak.position_uncertainty_2
    _nef_peak.position_3
    _nef_peak.position_uncertainty_3
    _nef_peak.height
    _nef_peak.height_uncertainty
    _nef_peak.volume
    _nef_peak.volume_uncertainty
    _nef_peak.ccpn_comment
    _nef_peak.ccpn_figure_of_merit


    0   0   A   1   MET   CA   A   2   GLN   N   A   2   GLN   H   54.608   .   122.899   .   8.907   .   .   .   .   .     merit=M   0.75
    1   1   A   2   GLN   CA   A   3   ILE   N   A   3   ILE   H   55.057   .   115.127   .   8.318   .   .   .   .   .     merit=M   0.75
    2   2   A   3   ILE   CA   A   4   PHE   N   A   4   PHE   H   59.599   .   118.73    .   8.614   .   .   .   .   .     merit=H   1.0
    3   3   A   4   PHE   CA   A   5   VAL   N   A   5   VAL   H   55.142   .   121.424   .   9.307   .   .   .   .   .     merit=H   1.0
stop_
""".replace(
    NOQUA_E501, ""
)


EXPECTED_all = """
# noqa E501
loop_
    _nef_peak.index
    _nef_peak.peak_id
    _nef_peak.chain_code_1
    _nef_peak.sequence_code_1
    _nef_peak.residue_name_1
    _nef_peak.atom_name_1
    _nef_peak.chain_code_2
    _nef_peak.sequence_code_2
    _nef_peak.residue_name_2
    _nef_peak.atom_name_2
    _nef_peak.chain_code_3
    _nef_peak.sequence_code_3
    _nef_peak.residue_name_3
    _nef_peak.atom_name_3
    _nef_peak.position_1
    _nef_peak.position_uncertainty_1
    _nef_peak.position_2
    _nef_peak.position_uncertainty_2
    _nef_peak.position_3
    _nef_peak.position_uncertainty_3
    _nef_peak.height
    _nef_peak.height_uncertainty
    _nef_peak.volume
    _nef_peak.volume_uncertainty
    _nef_peak.ccpn_comment
    _nef_peak.ccpn_figure_of_merit


    0   0   A   1   MET   CA   A   2   GLN   N   A   2   GLN   H   54.608   .   122.899   .   8.907   .   .   .   .   .      merit=M   0.75
    1   1   A   1   MET   CB   A   2   GLN   N   A   2   GLN   H   33.279   .   122.899   .   8.907   .   .   .   .   .      merit=M   0.75
    2   2   A   2   GLN   CA   A   2   GLN   N   A   2   GLN   H   55.215   .   122.899   .   8.907   .   .   .   .   .      merit=M   0.75
    3   3   A   2   GLN   CB   A   2   GLN   N   A   2   GLN   H   30.606   .   122.899   .   8.907   .   .   .   .   .      merit=M   0.75
    4   4   A   2   GLN   CA   A   3   ILE   N   A   3   ILE   H   55.057   .   115.127   .   8.318   .   .   .   .   .      merit=M   0.75
    5   5   A   2   GLN   CB   A   3   ILE   N   A   3   ILE   H   30.582   .   115.127   .   8.318   .   .   .   .   .      merit=M   0.75
    6   6   A   3   ILE   CA   A   3   ILE   N   A   3   ILE   H   59.616   .   115.127   .   8.318   .   .   .   .   .      merit=M   0.75
    7   7   A   3   ILE   CB   A   3   ILE   N   A   3   ILE   H   42.025   .   115.127   .   8.318   .   .   .   .   .      merit=M   0.75
stop_
""".replace(
    NOQUA_E501, ""
)


def test_import_peaks():

    sequence_stream = open(path_in_test_data(__file__, "ubi_seq.nef")).read()

    filenames = "sparky_all.out sparky_CA-1.out sparky_CA.out".split()

    paths = [path_in_test_data(__file__, filename) for filename in filenames]
    result = run_and_report(app, paths, input=sequence_stream)

    loop = isolate_loop(result.stdout, "nef_nmr_spectrum_mars_HNCA", "nef_peak")
    assert_lines_match(EXPECTED_HNCA, loop)

    loop = isolate_loop(result.stdout, "nef_nmr_spectrum_mars_HNcoCA", "nef_peak")
    assert_lines_match(EXPECTED_HNcoCA, loop)

    loop = isolate_loop(result.stdout, "nef_nmr_spectrum_mars_sparky_all", "nef_peak")
    assert_lines_match(EXPECTED_all, loop)



import pytest

from nef_pipelines.lib.sequence_lib import sequence_to_residue_name_lookup
from nef_pipelines.lib.structures import (
    AtomLabel,
    DimensionInfo,
    NewPeak,
    Residue,
    SequenceResidue,
    ShiftData,
)
from nef_pipelines.lib.test_lib import path_in_test_data
from nef_pipelines.transcoders.xeasy.xeasy_lib import (
    parse_peaks,
    parse_sequence,
    parse_shifts,
)

EXPECTED_SEQUENCE = [
    SequenceResidue(*residue.split())
    for residue in [
        "A 1 HIS",
        "A 2 MET",
        "A 3 ARG",
        "A 4 GLN",
        "A 5 THR",
        "A 6 MET",
        "A 7 LEU",
        "A 8 VAL",
        "A 9 THR",
    ]
]


def test_sequence_basic():
    sequence = open(path_in_test_data(__file__, "basic.seq")).readlines()

    result = parse_sequence(sequence)
    assert result == EXPECTED_SEQUENCE


def test_sequence_basic_commented():
    sequence = open(path_in_test_data(__file__, "basic.seq")).readlines()
    sequence = ["#a comment", *sequence]

    result = parse_sequence(sequence)
    assert result == EXPECTED_SEQUENCE


def test_sequence_basic_commented_two_lines_bad():
    sequence = open(path_in_test_data(__file__, "basic.seq")).readlines()

    with pytest.raises(SystemExit) as e:
        sequence = ["#a comment", "# a bad comment", *sequence]
        parse_sequence(sequence)

    assert e.type == SystemExit
    assert e.value.code == 1


EXPECTED_DIMENSIONS = [
    DimensionInfo(*dimension.split())
    for dimension in ["H1  HN ppm", "H1  H  ppm", "N15 N  ppm"]
]

EXPECTED_PEAKS = [
    NewPeak(
        [
            ShiftData(
                AtomLabel(
                    SequenceResidue(
                        "A",
                        "2",
                        "MET",
                    ),
                    "H",
                ),
                8.731,
            ),
            ShiftData(
                AtomLabel(
                    SequenceResidue(
                        "A",
                        "3",
                        "ARG",
                    ),
                    "H",
                ),
                8.723,
            ),
            ShiftData(
                AtomLabel(
                    SequenceResidue(
                        "A",
                        "4",
                        "GLN",
                    ),
                    "N",
                ),
                115.265,
            ),
        ],
        id=1,
        volume=5.5,
        volume_uncertainty=0.0,
        comment="MAP    403",
    ),
    NewPeak(
        [
            ShiftData(
                AtomLabel(
                    SequenceResidue(
                        ".",
                        ".",
                        ".",
                    ),
                    ".",
                ),
                8.731,
            ),
            ShiftData(
                AtomLabel(
                    SequenceResidue(
                        ".",
                        ".",
                        ".",
                    ),
                    ".",
                ),
                4.610,
            ),
            ShiftData(
                AtomLabel(
                    SequenceResidue(
                        ".",
                        ".",
                        ".",
                    ),
                    ".",
                ),
                115.275,
            ),
        ],
        id=2,
        volume=5.5,
        volume_uncertainty=0.0,
        comment=None,
    ),
    NewPeak(
        [
            ShiftData(
                AtomLabel(
                    SequenceResidue(
                        "A",
                        "5",
                        "THR",
                    ),
                    "H",
                ),
                8.732,
            ),
            ShiftData(
                AtomLabel(
                    SequenceResidue(
                        "A",
                        "6",
                        "MET",
                    ),
                    "HA2",
                ),
                4.192,
            ),
            ShiftData(
                AtomLabel(
                    SequenceResidue(
                        "A",
                        "7",
                        "LEU",
                    ),
                    "N",
                ),
                115.254,
            ),
        ],
        id=3,
        volume=5.5,
        volume_uncertainty=0.0,
        comment="MAP    404",
    ),
    NewPeak(
        [
            ShiftData(
                AtomLabel(
                    SequenceResidue(
                        "A",
                        "5",
                        "THR",
                    ),
                    "H",
                ),
                8.732,
            ),
            ShiftData(
                AtomLabel(
                    SequenceResidue(
                        "A",
                        "6",
                        "MET",
                    ),
                    "HA3",
                ),
                4.192,
            ),
            ShiftData(
                AtomLabel(
                    SequenceResidue(
                        "A",
                        "7",
                        "LEU",
                    ),
                    "N",
                ),
                115.254,
            ),
        ],
        id=3,
        volume=5.5,
        volume_uncertainty=0.0,
        comment="MAP    405",
    ),
]


def test_basic_peaks():
    peaks = open(path_in_test_data(__file__, "basic.peaks")).readlines()

    lookup = sequence_to_residue_name_lookup(EXPECTED_SEQUENCE)
    spectrum_type, dimension_info, peaks = parse_peaks(
        peaks, source="unknown", residue_name_lookup=lookup
    )

    assert spectrum_type == "N15TOCSY"
    assert dimension_info == EXPECTED_DIMENSIONS

    assert peaks == EXPECTED_PEAKS


EXPECTED_SHIFTS = [
    ShiftData(
        atom=AtomLabel(
            residue=Residue(chain_code="A", sequence_code="1", residue_name="HIS"),
            atom_name="N",
            element=None,
            isotope_number=None,
        ),
        value="112.794",
        value_uncertainty="0.010",
        line_width=None,
        line_width_uncertainty=None,
    ),
    ShiftData(
        atom=AtomLabel(
            residue=Residue(chain_code="A", sequence_code="1", residue_name="HIS"),
            atom_name="HE1",
            element=None,
            isotope_number=None,
        ),
        value="7.955",
        value_uncertainty="0.010",
        line_width=None,
        line_width_uncertainty=None,
    ),
    ShiftData(
        atom=AtomLabel(
            residue=Residue(chain_code="A", sequence_code="2", residue_name="MET"),
            atom_name="N",
            element=None,
            isotope_number=None,
        ),
        value="113.996",
        value_uncertainty="0.000",
        line_width=None,
        line_width_uncertainty=None,
    ),
    ShiftData(
        atom=AtomLabel(
            residue=Residue(chain_code="A", sequence_code="2", residue_name="MET"),
            atom_name="QE",
            element=None,
            isotope_number=None,
        ),
        value="3.778",
        value_uncertainty="0.010",
        line_width=None,
        line_width_uncertainty=None,
    ),
    ShiftData(
        atom=AtomLabel(
            residue=Residue(chain_code="A", sequence_code="2", residue_name="MET"),
            atom_name="CE",
            element=None,
            isotope_number=None,
        ),
        value="27.054",
        value_uncertainty="0.010",
        line_width=None,
        line_width_uncertainty=None,
    ),
    ShiftData(
        atom=AtomLabel(
            residue=Residue(chain_code="A", sequence_code="3", residue_name="ARG"),
            atom_name="HD2",
            element=None,
            isotope_number=None,
        ),
        value="3.133",
        value_uncertainty="0.000",
        line_width=None,
        line_width_uncertainty=None,
    ),
    ShiftData(
        atom=AtomLabel(
            residue=Residue(chain_code="A", sequence_code="3", residue_name="ARG"),
            atom_name="HD3",
            element=None,
            isotope_number=None,
        ),
        value="3.134",
        value_uncertainty="0.000",
        line_width=None,
        line_width_uncertainty=None,
    ),
    ShiftData(
        atom=AtomLabel(
            residue=Residue(chain_code="A", sequence_code="3", residue_name="ARG"),
            atom_name="HE",
            element=None,
            isotope_number=None,
        ),
        value="4.378",
        value_uncertainty="0.010",
        line_width=None,
        line_width_uncertainty=None,
    ),
]


def test_basic_shifts():

    shifts = open(path_in_test_data(__file__, "basic_shifts.prot")).readlines()

    lookup = sequence_to_residue_name_lookup(EXPECTED_SEQUENCE)
    shifts = parse_shifts(shifts, source="unknown", residue_lookup=lookup)

    assert shifts == EXPECTED_SHIFTS
